"""Tests for task optimization."""
import dspy
import pydantic
import pytest

from sieves import ModelSettings
from sieves.model_wrappers import ModelType
from sieves.tasks.optimization import Optimizer
from sieves.tasks.predictive import schemas
from sieves.tasks.predictive import (
    classification,
    sentiment_analysis,
    ner,
    pii_masking,
    information_extraction,
    summarization,
    translation,
    question_answering,
)

from sieves.tests.conftest import make_model


@pytest.fixture(scope="module")
def optimizer(request) -> Optimizer:
    """Return model and optimizer to use for optimization.

    :return: model and optimizer to use for optimization.
    """
    model = make_model(ModelType.dspy)
    optimizer = Optimizer(
        model,
        val_frac=.25,
        shuffle=True,
        dspy_init_kwargs=dict(
            auto=None, num_candidates=1, max_bootstrapped_demos=1, max_labeled_demos=1, max_errors=10, num_threads=1
        ),
        dspy_compile_kwargs=dict(num_trials=1, minibatch=False),
    )

    return optimizer


def test_optimization_classification(optimizer) -> None:
    """Tests optimization for classification tasks."""
    examples_single_label = [
        classification.FewshotExampleSingleLabel(text='Apple', label='fruit', score=1.),
        classification.FewshotExampleSingleLabel(text='Broccoli', label='vegetable', score=1.),
        classification.FewshotExampleSingleLabel(text='Melon', label='fruit', score=1.),
        classification.FewshotExampleSingleLabel(text='Carrot', label='vegetable', score=1.),
        classification.FewshotExampleSingleLabel(text='Tomato', label='fruit', score=1.),
        classification.FewshotExampleSingleLabel(text='Pepper', label='vegetable',
                                                 score=1.),
        classification.FewshotExampleSingleLabel(text='Kiwi', label='fruit', score=1.),
        classification.FewshotExampleSingleLabel(text='Onion', label='vegetable',
                                                 score=1.),
    ]
    examples_multi_label = [
        classification.FewshotExampleMultiLabel(
            text='Ghostbusters',
            score_per_label={'comedy': 0.9, 'scifi': 0.8}
        ),
        classification.FewshotExampleMultiLabel(
            text='The Martian',
            score_per_label={'comedy': 0.4, 'scifi': 1.0}
        ),
        classification.FewshotExampleMultiLabel(
            text='Galaxy Quest',
            score_per_label={'comedy': 1.0, 'scifi': 0.9}
        ),
        classification.FewshotExampleMultiLabel(
            text='Back to the Future',
            score_per_label={'comedy': 0.8, 'scifi': 0.9}
        ),
        classification.FewshotExampleMultiLabel(
            text='Superbad',
            score_per_label={'comedy': 1.0, 'scifi': 0.0}
        ),
        classification.FewshotExampleMultiLabel(
            text='Blade Runner 2049',
            score_per_label={'comedy': 0.05, 'scifi': 1.0}
        ),
        classification.FewshotExampleMultiLabel(
            text='Guardians of the Galaxy',
            score_per_label={'comedy': 0.75, 'scifi': 0.9}
        ),
        classification.FewshotExampleMultiLabel(
            text='Interstellar',
            score_per_label={'comedy': 0.05, 'scifi': 1.0}
        ),
    ]

    task_single_label = classification.Classification(
        mode='single',
        labels=["fruit", "vegetable"],
        fewshot_examples=examples_single_label,
        model=optimizer.model,
        model_settings=ModelSettings(),
    )
    task_multi_label = classification.Classification(
        mode='multi',
        labels=["comedy", "scifi"],
        fewshot_examples=examples_multi_label,
        model=optimizer.model,
        model_settings=ModelSettings(),
    )

    # Test evaluation.
    assert task_single_label._evaluate_dspy_example(
        truth=dspy.Example(text="", label="fruit", score=.7),
        pred=dspy.Prediction(text="", label="fruit", score=.1),
        trace=None,
        model=optimizer.model,
    ) == 1.
    assert task_single_label._evaluate_dspy_example(
        truth=dspy.Example(text="", label="fruit", score=.7),
        pred=dspy.Prediction(text="", label="vegetable", score=.1),
        trace=None,
        model=optimizer.model,
    ) == 0.
    assert task_multi_label._evaluate_dspy_example(
        truth=dspy.Example(text="", score_per_label={"comedy": .4, "scifi": .2}),
        pred=dspy.Prediction(text="", score_per_label={"comedy": .1, "scifi": .3}),
        trace=None,
        model=optimizer.model,
    ) == (0.7 + 0.9) / 2
    assert task_multi_label._evaluate_dspy_example(
        truth=dspy.Example(text="", score_per_label={"comedy": .4, "scifi": .2}),
        pred=dspy.Prediction(text="", score_per_label={"comedy": .4, "scifi": .2}),
        trace=None,
        model=optimizer.model,
    ) == 1

    # Smoke-test optimization.
    best_prompt, best_examples = task_single_label.optimize(optimizer, verbose=False)
    assert task_single_label._custom_prompt_instructions == best_prompt
    assert task_single_label._bridge._prompt_instructions == best_prompt
    assert isinstance(task_single_label._fewshot_examples, list)

    best_prompt, best_examples = task_multi_label.optimize(optimizer, verbose=False)
    assert task_multi_label._custom_prompt_instructions == best_prompt
    assert task_multi_label._bridge._prompt_instructions == best_prompt
    assert isinstance(task_multi_label._fewshot_examples, list)


def test_optimization_sentiment_analysis(optimizer) -> None:
    """Tests optimization for sentiment analysis task."""
    examples = [
        sentiment_analysis.FewshotExample(
            text='Great product, excellent quality and fast shipping!',
            sentiment_per_aspect={'overall': 0.95, 'quality': 0.9, 'delivery': 0.95}
        ),
        sentiment_analysis.FewshotExample(
            text='Terrible quality, arrived damaged and late.',
            sentiment_per_aspect={'overall': 0.1, 'quality': 0.05, 'delivery': 0.15}
        ),
        sentiment_analysis.FewshotExample(
            text='Decent product but shipping was slow.',
            sentiment_per_aspect={'overall': 0.5, 'quality': 0.6, 'delivery': 0.3}
        ),
        sentiment_analysis.FewshotExample(
            text='Amazing quality! Worth every penny.',
            sentiment_per_aspect={'overall': 0.95, 'quality': 1.0, 'delivery': 0.5}
        ),
        sentiment_analysis.FewshotExample(
            text='Not great, not terrible. Just okay.',
            sentiment_per_aspect={'overall': 0.5, 'quality': 0.5, 'delivery': 0.5}
        ),
        sentiment_analysis.FewshotExample(
            text='Quick delivery but product quality is poor.',
            sentiment_per_aspect={'overall': 0.4, 'quality': 0.2, 'delivery': 0.8}
        ),
    ]

    task = sentiment_analysis.SentimentAnalysis(
        aspects=('quality', 'delivery'),
        model=optimizer.model,
        fewshot_examples=examples,
        model_settings=ModelSettings(),
    )

    # Test evaluation: perfect match
    assert task._evaluate_dspy_example(
        truth=dspy.Example(text='', reasoning='', sentiment_per_aspect={'overall': 0.8, 'quality': 0.7, 'delivery': 0.9}),
        pred=dspy.Prediction(text='', reasoning='', sentiment_per_aspect={'overall': 0.8, 'quality': 0.7, 'delivery': 0.9}),
        trace=None,
        model=optimizer.model
    ) == 1.0

    # Test evaluation: partial match (MAE-based accuracy)
    # |0.8-0.6| = 0.2, |0.7-0.5| = 0.2, |0.9-0.8| = 0.1
    # Accuracy per aspect: (1-0.2) + (1-0.2) + (1-0.1) = 0.8 + 0.8 + 0.9 = 2.5
    # Average: 2.5 / 3 ≈ 0.833
    score = task._evaluate_dspy_example(
        truth=dspy.Example(text='', reasoning='', sentiment_per_aspect={'overall': 0.8, 'quality': 0.7, 'delivery': 0.9}),
        pred=dspy.Prediction(text='', reasoning='', sentiment_per_aspect={'overall': 0.6, 'quality': 0.5, 'delivery': 0.8}),
        trace=None,
        model=optimizer.model,
    )
    assert abs(score - 0.833) < 0.01

    # Smoke-test optimization
    best_prompt, best_examples = task.optimize(optimizer, verbose=False)
    assert task._custom_prompt_instructions == best_prompt
    assert task._bridge._prompt_instructions == best_prompt
    assert isinstance(task._fewshot_examples, list)


def test_optimization_ner(optimizer) -> None:
    """Tests optimization for NER task."""
    examples = [
        ner.FewshotExample(
            text='John Smith visited Paris last week.',
            entities=[
                schemas.ner.EntityWithContext(text='John Smith', context='visited Paris', entity_type='PERSON'),
                schemas.ner.EntityWithContext(text='Paris', context='John Smith visited', entity_type='LOCATION'),
            ]
        ),
        ner.FewshotExample(
            text='Apple CEO Tim Cook announced new products.',
            entities=[
                schemas.ner.EntityWithContext(text='Tim Cook', context='Apple CEO', entity_type='PERSON'),
            ]
        ),
        ner.FewshotExample(
            text='The meeting in London was attended by Sarah Johnson.',
            entities=[
                schemas.ner.EntityWithContext(text='London', context='meeting in', entity_type='LOCATION'),
                schemas.ner.EntityWithContext(text='Sarah Johnson', context='attended by', entity_type='PERSON'),
            ]
        ),
        ner.FewshotExample(
            text='Berlin is the capital of Germany.',
            entities=[
                schemas.ner.EntityWithContext(text='Berlin', context='capital of Germany', entity_type='LOCATION'),
            ]
        ),
        ner.FewshotExample(
            text='Maria Rodriguez traveled to Tokyo.',
            entities=[
                schemas.ner.EntityWithContext(
                    text='Maria Rodriguez', context='traveled to Tokyo', entity_type='PERSON'
                ),
                schemas.ner.EntityWithContext(text='Tokyo', context='Maria Rodriguez traveled', entity_type='LOCATION'),
            ]
        ),
        ner.FewshotExample(
            text='The conference in New York was successful.',
            entities=[
                schemas.ner.EntityWithContext(text='New York', context='conference in', entity_type='LOCATION'),
            ]
        ),
    ]

    task = ner.NER(
        entities=['PERSON', 'LOCATION'],
        model=optimizer.model,
        fewshot_examples=examples,
        model_settings=ModelSettings(),
    )

    # Test evaluation: perfect match (F1 = 1.0)
    assert task._evaluate_dspy_example(
        truth=dspy.Example(text='', entities=[
            {'text': 'Alice', 'entity_type': 'PERSON'},
            {'text': 'Boston', 'entity_type': 'LOCATION'},
        ]),
        pred=dspy.Prediction(text='', entities=[
            {'text': 'Alice', 'entity_type': 'PERSON'},
            {'text': 'Boston', 'entity_type': 'LOCATION'},
        ]),
        trace=None,
        model=optimizer.model,
    ) == 1.0

    # Test evaluation: partial match (precision=0.5, recall=0.5, F1=0.5)
    # True: {('Alice', 'PERSON'), ('Boston', 'LOCATION')}
    # Pred: {('Alice', 'PERSON'), ('Chicago', 'LOCATION')}
    # Intersection: {('Alice', 'PERSON')} = 1
    # Precision: 1/2 = 0.5, Recall: 1/2 = 0.5, F1: 0.5
    assert task._evaluate_dspy_example(
        truth=dspy.Example(text='', entities=[
            {'text': 'Alice', 'entity_type': 'PERSON'},
            {'text': 'Boston', 'entity_type': 'LOCATION'},
        ]),
        pred=dspy.Prediction(text='', entities=[
            {'text': 'Alice', 'entity_type': 'PERSON'},
            {'text': 'Chicago', 'entity_type': 'LOCATION'},
        ]),
        trace=None,
        model=optimizer.model
    ) == 0.5

    # Test evaluation: no entities in truth (empty case)
    assert task._evaluate_dspy_example(
        truth=dspy.Example(text='', entities=[]),
        pred=dspy.Prediction(text='', entities=[]),
        trace=None,
        model=optimizer.model
    ) == 1.0

    # Smoke-test optimization
    best_prompt, best_examples = task.optimize(optimizer, verbose=False)
    assert task._custom_prompt_instructions == best_prompt
    assert task._bridge._prompt_instructions == best_prompt
    assert isinstance(task._fewshot_examples, list)


def test_optimization_pii_masking(optimizer) -> None:
    """Tests optimization for PII masking task."""
    examples = [
        pii_masking.FewshotExample(
            text='Please contact John Doe at john.doe@email.com for more information.',
            masked_text='Please contact [MASKED] at [MASKED] for more information.',
            pii_entities=[
                pii_masking.PIIEntity(entity_type='NAME', text='John Doe'),
                pii_masking.PIIEntity(entity_type='EMAIL', text='john.doe@email.com'),
            ]
        ),
        pii_masking.FewshotExample(
            text='Send the report to alice.smith@company.com by Friday.',
            masked_text='Send the report to [MASKED] by Friday.',
            pii_entities=[
                pii_masking.PIIEntity(entity_type='EMAIL', text='alice.smith@company.com'),
            ]
        ),
        pii_masking.FewshotExample(
            text='Call Bob Johnson at the office tomorrow.',
            masked_text='Call [MASKED] at the office tomorrow.',
            pii_entities=[
                pii_masking.PIIEntity(entity_type='NAME', text='Bob Johnson'),
            ]
        ),
        pii_masking.FewshotExample(
            text='Sarah Miller will attend the meeting.',
            masked_text='[MASKED] will attend the meeting.',
            pii_entities=[
                pii_masking.PIIEntity(entity_type='NAME', text='Sarah Miller'),
            ]
        ),
        pii_masking.FewshotExample(
            text='Email the document to michael.brown@org.net and copy jane.white@org.net.',
            masked_text='Email the document to [MASKED] and copy [MASKED].',
            pii_entities=[
                pii_masking.PIIEntity(entity_type='EMAIL', text='michael.brown@org.net'),
                pii_masking.PIIEntity(entity_type='EMAIL', text='jane.white@org.net'),
            ]
        ),
        pii_masking.FewshotExample(
            text='The meeting is scheduled for 2pm.',
            masked_text='The meeting is scheduled for 2pm.',
            pii_entities=[]
        ),
    ]

    task = pii_masking.PIIMasking(
        pii_types=['NAME', 'EMAIL'],
        model=optimizer.model,
        fewshot_examples=examples,
        model_settings=ModelSettings(),
    )

    # Test evaluation: perfect match (F1 = 1.0)
    assert task._evaluate_dspy_example(
        truth=dspy.Example(text='', reasoning='', masked_text='', pii_entities=[
            {'entity_type': 'NAME', 'text': 'Alice'},
            {'entity_type': 'EMAIL', 'text': 'alice@test.com'},
        ]),
        pred=dspy.Prediction(text='', reasoning='', masked_text='', pii_entities=[
            {'entity_type': 'NAME', 'text': 'Alice'},
            {'entity_type': 'EMAIL', 'text': 'alice@test.com'},
        ]),
        trace=None,
model=optimizer.model
    ) == 1.0

    # Test evaluation: partial match (precision=0.5, recall=0.5, F1=0.5)
    # True: {('NAME', 'Alice'), ('EMAIL', 'alice@test.com')}
    # Pred: {('NAME', 'Alice'), ('EMAIL', 'bob@test.com')}
    # Intersection: {('NAME', 'Alice')} = 1
    # Precision: 1/2 = 0.5, Recall: 1/2 = 0.5, F1: 0.5
    assert task._evaluate_dspy_example(
        truth=dspy.Example(text='', reasoning='', masked_text='', pii_entities=[
            {'entity_type': 'NAME', 'text': 'Alice'},
            {'entity_type': 'EMAIL', 'text': 'alice@test.com'},
        ]),
        pred=dspy.Prediction(text='', reasoning='', masked_text='', pii_entities=[
            {'entity_type': 'NAME', 'text': 'Alice'},
            {'entity_type': 'EMAIL', 'text': 'bob@test.com'},
        ]),
        trace=None,
model=optimizer.model
    ) == 0.5

    # Test evaluation: no PII (empty case)
    assert task._evaluate_dspy_example(
        truth=dspy.Example(text='', reasoning='', masked_text='', pii_entities=[]),
        pred=dspy.Prediction(text='', reasoning='', masked_text='', pii_entities=[]),
        trace=None,
model=optimizer.model
    ) == 1.0

    # Smoke-test optimization
    best_prompt, best_examples = task.optimize(optimizer, verbose=False)
    assert task._custom_prompt_instructions == best_prompt
    assert task._bridge._prompt_instructions == best_prompt
    assert isinstance(task._fewshot_examples, list)


def test_optimization_information_extraction(optimizer) -> None:
    """Tests optimization for information extraction task."""
    # Define entity type for extraction
    class Person(pydantic.BaseModel, frozen=True):
        """Person entity."""
        name: str
        age: int
        occupation: str

    examples = [
        information_extraction.FewshotExampleMulti(
            text='Alice Johnson is a 28-year-old software engineer.',
            entities=[Person(name='Alice Johnson', age=28, occupation='software engineer')]
        ),
        information_extraction.FewshotExampleMulti(
            text='Bob Smith, age 35, works as a teacher.',
            entities=[Person(name='Bob Smith', age=35, occupation='teacher')]
        ),
        information_extraction.FewshotExampleMulti(
            text='The team includes Sarah Lee (42, doctor) and Mike Brown (30, lawyer).',
            entities=[
                Person(name='Sarah Lee', age=42, occupation='doctor'),
                Person(name='Mike Brown', age=30, occupation='lawyer'),
            ]
        ),
        information_extraction.FewshotExampleMulti(
            text='Emma Davis is 25 and works as a designer.',
            entities=[Person(name='Emma Davis', age=25, occupation='designer')]
        ),
        information_extraction.FewshotExampleMulti(
            text='Dr. John Williams, a 50-year-old researcher, published a new paper.',
            entities=[Person(name='Dr. John Williams', age=50, occupation='researcher')]
        ),
        information_extraction.FewshotExampleMulti(
            text='The company hired Lisa Chen (27) as an analyst.',
            entities=[Person(name='Lisa Chen', age=27, occupation='analyst')]
        ),
    ]

    task = information_extraction.InformationExtraction(
        entity_type=Person,
        model=optimizer.model,
        fewshot_examples=examples,
        model_settings=ModelSettings(),
    )

    # Test evaluation: perfect match (F1 = 1.0)
    assert task._evaluate_dspy_example(
        truth=dspy.Example(text='', reasoning='', entities=[
            {'name': 'Alice', 'age': 30, 'occupation': 'engineer'},
            {'name': 'Bob', 'age': 25, 'occupation': 'designer'},
        ]),
        pred=dspy.Prediction(text='', reasoning='', entities=[
            {'name': 'Alice', 'age': 30, 'occupation': 'engineer'},
            {'name': 'Bob', 'age': 25, 'occupation': 'designer'},
        ]),
        trace=None,
model=optimizer.model
    ) == 1.0

    # Test evaluation: partial match (precision=0.5, recall=0.5, F1=0.5)
    # True entities (as sorted tuples): {('age',25),('name','Alice'),('occupation','engineer'), ('age',30),
    # ('name','Bob'),('occupation','designer')}
    # Pred entities: one correct, one different
    # This should give F1 of 0.5
    assert task._evaluate_dspy_example(
        truth=dspy.Example(text='', reasoning='', entities=[
            {'name': 'Alice', 'age': 30, 'occupation': 'engineer'},
            {'name': 'Bob', 'age': 25, 'occupation': 'designer'},
        ]),
        pred=dspy.Prediction(text='', reasoning='', entities=[
            {'name': 'Alice', 'age': 30, 'occupation': 'engineer'},
            {'name': 'Charlie', 'age': 35, 'occupation': 'teacher'},
        ]),
        trace=None,
model=optimizer.model
    ) == 0.5

    # Test evaluation: no entities (empty case)
    assert task._evaluate_dspy_example(
        truth=dspy.Example(text='', reasoning='', entities=[]),
        pred=dspy.Prediction(text='', reasoning='', entities=[]),
        trace=None,
model=optimizer.model
    ) == 1.0

    # Smoke-test optimization
    best_prompt, best_examples = task.optimize(optimizer, verbose=False)
    assert task._custom_prompt_instructions == best_prompt
    assert task._bridge._prompt_instructions == best_prompt
    assert isinstance(task._fewshot_examples, list)


def test_optimization_summarization(optimizer) -> None:
    """Tests optimization for summarization task using LLM-based evaluator."""
    examples = [
        summarization.FewshotExample(
            text='The European Space Agency launched a new satellite yesterday to monitor climate change. '
                 'The satellite will collect data on temperature, sea levels, and atmospheric conditions over the '
                 'next decade.',
            n_words=30,
            summary='ESA launched a climate monitoring satellite to track temperature, sea levels, and atmosphere for '
                    '10 years.'
        ),
        summarization.FewshotExample(
            text='Local farmers are adopting sustainable practices to reduce water usage and improve soil health. '
                 'New irrigation systems have reduced water consumption by 40% while maintaining crop yields.',
            n_words=30,
            summary='Farmers use sustainable methods and new irrigation, cutting water use 40% with same yields.'
        ),
        summarization.FewshotExample(
            text='The company announced record profits of $5 billion this quarter, driven by strong sales in Asia '
                 'and Europe. CEO Jane Smith credited the success to innovative product launches and effective '
                 'marketing.',
            n_words=30,
            summary='Company posts $5B profit from Asian/European sales, CEO cites innovation and marketing.'
        ),
        summarization.FewshotExample(
            text='Researchers discovered a new species of deep-sea fish near the Mariana Trench. '
                 'The bioluminescent creature can survive at depths of 8,000 meters and uses light to attract prey.',
            n_words=30,
            summary='Scientists find new bioluminescent deep-sea fish at 8,000m depth that uses light for hunting.'
        ),
        summarization.FewshotExample(
            text='The city council approved a $200 million infrastructure plan to repair roads, bridges, and public transit. '
                 'Construction will begin next spring and is expected to create 5,000 jobs over three years.',
            n_words=30,
            summary='Council approves $200M infrastructure plan for roads, bridges, transit; 5,000 jobs starting spring.'
        ),
        summarization.FewshotExample(
            text='Scientists developed a new vaccine that shows 95% effectiveness against multiple virus strains. '
                 'Clinical trials involved 50,000 participants across 20 countries and demonstrated strong safety profiles.',
            n_words=30,
            summary='New vaccine shows 95% effectiveness across virus strains in 50,000-participant global trials.'
        ),
    ]

    task = summarization.Summarization(
        n_words=30,
        model=optimizer.model,
        fewshot_examples=examples,
        model_settings=ModelSettings(),
    )

    # Test LLM-based evaluation (no hardcoded scores since it uses LLM)
    # Just verify it runs and returns a score between 0 and 1
    score = task._evaluate_dspy_example(
        truth=dspy.Example(text='', n_words=30, summary='Short summary about climate.'),
        pred=dspy.Prediction(text='', n_words=30, summary='Brief summary on climate change.'),
        trace=None,
model=optimizer.model
    )
    assert 0.5 <= score <= 1.0

    # Smoke-test optimization
    best_prompt, best_examples = task.optimize(optimizer, verbose=False)
    assert task._custom_prompt_instructions == best_prompt
    assert task._bridge._prompt_instructions == best_prompt
    assert isinstance(task._fewshot_examples, list)


def test_optimization_translation(optimizer) -> None:
    """Tests optimization for translation task using LLM-based evaluator."""
    examples = [
        translation.FewshotExample(
            text='Hello, how are you today?',
            to='Spanish',
            translation='Hola, ¿cómo estás hoy?'
        ),
        translation.FewshotExample(
            text='The weather is beautiful this morning.',
            to='Spanish',
            translation='El clima está hermoso esta mañana.'
        ),
        translation.FewshotExample(
            text='I would like to order a coffee, please.',
            to='Spanish',
            translation='Me gustaría pedir un café, por favor.'
        ),
        translation.FewshotExample(
            text='Where is the nearest train station?',
            to='Spanish',
            translation='¿Dónde está la estación de tren más cercana?'
        ),
        translation.FewshotExample(
            text='Thank you very much for your help.',
            to='Spanish',
            translation='Muchas gracias por tu ayuda.'
        ),
        translation.FewshotExample(
            text='I am learning Spanish and enjoying it.',
            to='Spanish',
            translation='Estoy aprendiendo español y lo disfruto.'
        ),
    ]

    task = translation.Translation(
        to='Spanish',
        model=optimizer.model,
        fewshot_examples=examples,
        model_settings=ModelSettings(),
    )

    # Test LLM-based evaluation
    score = task._evaluate_dspy_example(
        truth=dspy.Example(text='Good morning.', to='Spanish', translation='Buenos días.'),
        pred=dspy.Prediction(text='Good morning.', to='Spanish', translation='Buen día.'),
        trace=None,
model=optimizer.model
    )
    assert 0.7 <= score <= 1.0

    # Smoke-test optimization
    best_prompt, best_examples = task.optimize(optimizer, verbose=False)
    assert task._custom_prompt_instructions == best_prompt
    assert task._bridge._prompt_instructions == best_prompt
    assert isinstance(task._fewshot_examples, list)


def test_optimization_question_answering(optimizer) -> None:
    """Tests optimization for question answering task using LLM-based evaluator."""
    questions = ['What is the main topic?', 'Who are the key people mentioned?']

    examples = [
        question_answering.FewshotExample(
            text='Albert Einstein developed the theory of relativity in 1915. His work revolutionized modern physics.',
            questions=questions,
            answers=['The theory of relativity', 'Albert Einstein']
        ),
        question_answering.FewshotExample(
            text='Marie Curie won two Nobel Prizes for her research on radioactivity. She was the first woman to win a Nobel Prize.',
            questions=questions,
            answers=['Research on radioactivity and Nobel Prizes', 'Marie Curie']
        ),
        question_answering.FewshotExample(
            text='Shakespeare wrote Romeo and Juliet in 1597. The play is one of the most famous love stories in '
                 'literature.',
            questions=questions,
            answers=['Romeo and Juliet play', 'Shakespeare']
        ),
        question_answering.FewshotExample(
            text='The Amazon rainforest is home to millions of species. Deforestation threatens this biodiversity.',
            questions=questions,
            answers=['Amazon rainforest biodiversity and deforestation', 'No specific people mentioned']
        ),
        question_answering.FewshotExample(
            text='Neil Armstrong became the first person to walk on the moon in 1969. This historic event was part of '
                 'the Apollo 11 mission.',
            questions=questions,
            answers=['First moon landing', 'Neil Armstrong']
        ),
        question_answering.FewshotExample(
            text='Leonardo da Vinci painted the Mona Lisa during the Renaissance. He was also an inventor and scientist.',
            questions=questions,
            answers=['The Mona Lisa painting', 'Leonardo da Vinci']
        ),
    ]

    task = question_answering.QuestionAnswering(
        questions=questions,
        model=optimizer.model,
        fewshot_examples=examples,
        model_settings=ModelSettings(),
    )

    # Test LLM-based evaluation
    score = task._evaluate_dspy_example(
        truth=dspy.Example(
            text='',
            reasoning='',
            questions=questions,
            answers=['Climate change', 'Scientists'],
            scores=None
        ),
        pred=dspy.Prediction(
            text='',
            reasoning='',
            questions=questions,
            answers=['Global warming', 'Researchers'],
            scores=None
        ),
        trace=None,
        model=optimizer.model
    )
    assert 0.5 <= score <= 1.0

    # Smoke-test optimization
    best_prompt, best_examples = task.optimize(optimizer, verbose=False)
    assert task._custom_prompt_instructions == best_prompt
    assert task._bridge._prompt_instructions == best_prompt
    assert isinstance(task._fewshot_examples, list)



"""Import optional ingestion backends with graceful fallbacks.

If a library can't be found, a placeholder task is exported instead.This allows
downstream imports to succeed without hard dependencies.
"""

try:
    from sieves.tasks.preprocessing.ingestion import docling_  # type: ignore
    from sieves.tasks.preprocessing.ingestion.docling_ import Docling

except (ModuleNotFoundError, ImportError):
    from sieves.tasks.preprocessing.ingestion import missing as docling_  # type: ignore
    from sieves.tasks.preprocessing.ingestion.missing import MissingIngestion as Docling  # type: ignore


try:
    from sieves.tasks.preprocessing.ingestion import marker_  # type: ignore
    from sieves.tasks.preprocessing.ingestion.marker_ import Marker

except (ModuleNotFoundError, ImportError):
    from sieves.tasks.preprocessing.ingestion import missing as marker_  # type: ignore
    from sieves.tasks.preprocessing.ingestion.missing import MissingIngestion as Marker  # type: ignore


__all__ = [
    "docling_",
    "Docling",
    "marker_",
    "Marker",
]



# mypy: ignore-errors
from typing import Literal, Union, Any

import dspy
import gliner2.inference.engine
import pydantic
import pytest

from sieves.model_wrappers import ModelType
from sieves.tasks.predictive.utils import convert_to_signature

# --- Helper for comparison (reused from test_signature_utils.py) ---

def assert_dspy_sigs_equal(actual: type[dspy.Signature], expected: type[dspy.Signature]):
    assert actual.input_fields.keys() == expected.input_fields.keys()
    assert actual.output_fields.keys() == expected.output_fields.keys()

    for k in actual.input_fields:
        assert actual.input_fields[k].json_schema_extra.get("desc") == expected.input_fields[k].json_schema_extra.get("desc")

    for k in actual.output_fields:
        actual_desc = actual.output_fields[k].json_schema_extra.get("desc")
        expected_desc = expected.output_fields[k].json_schema_extra.get("desc")
        assert actual_desc == expected_desc

    assert actual.__doc__ == expected.__doc__

def assert_gliner_schemas_equal(actual: Any, expected: Any):
    if hasattr(actual, "_field_metadata"):
        assert actual._field_metadata == expected._field_metadata
    elif hasattr(actual, "schema"):
        if hasattr(actual.schema, "_field_metadata"):
            assert actual.schema._field_metadata == expected.schema._field_metadata
        else:
            assert actual.schema == expected.schema
    else:
        assert actual == expected

# --- Task-specific Tests ---

def test_classification_migration():
    # 1. Define Unified Pydantic Prompt Signature
    class ClassificationOutput(pydantic.BaseModel):
        """Classify the text into one of the provided labels."""
        label: Literal["science", "politics", "sports"] = pydantic.Field(description="The predicted label")
        score: float = pydantic.Field(description="Confidence score")

    # 2. Define Expected Output Signatures

    # DSPy
    class ExpectedDSPy(dspy.Signature):
        """Classify the text into one of the provided labels."""
        text = dspy.InputField(desc="Input text to process.")
        label = dspy.OutputField(desc="The predicted label")
        score = dspy.OutputField(desc="Confidence score")

    # GliNER (classification mode)
    expected_gliner = gliner2.inference.engine.Schema().classification(
        task="classification",
        labels=["science", "politics", "sports"]
    )

    # HuggingFace (labels)
    expected_hf = ["science", "politics", "sports"]

    # Outlines (choice mode)
    expected_outlines_choice = ["science", "politics", "sports"]

    # 3. Verify
    assert_dspy_sigs_equal(convert_to_signature(ClassificationOutput, ModelType.dspy), ExpectedDSPy)
    assert_gliner_schemas_equal(convert_to_signature(ClassificationOutput, ModelType.gliner, mode="classification"), expected_gliner)
    assert convert_to_signature(ClassificationOutput, ModelType.huggingface) == expected_hf
    assert convert_to_signature(ClassificationOutput, ModelType.outlines, inference_mode="choice") == expected_outlines_choice


def test_classification_multi_label_migration():
    # 1. Define Unified Pydantic Prompt Signature
    class MultiLabelClassificationOutput(pydantic.BaseModel):
        """Perform multi-label classification."""
        science: float = pydantic.Field(description="Score for science category")
        politics: float = pydantic.Field(description="Score for politics category")
        sports: float = pydantic.Field(description="Score for sports category")

    # 2. Define Expected Output Signatures

    # DSPy (Unified schema leads to flat output fields)
    class ExpectedDSPy(dspy.Signature):
        """Perform multi-label classification."""
        text = dspy.InputField(desc="Input text to process.")
        science = dspy.OutputField(desc="Score for science category")
        politics = dspy.OutputField(desc="Score for politics category")
        sports = dspy.OutputField(desc="Score for sports category")

    # GliNER (classification mode)
    expected_gliner = gliner2.inference.engine.Schema().classification(
        task="classification",
        labels=["science", "politics", "sports"]
    )

    # HuggingFace (labels)
    expected_hf = ["science", "politics", "sports"]

    # Outlines (JSON mode)
    expected_outlines = MultiLabelClassificationOutput

    # 3. Verify
    assert_dspy_sigs_equal(convert_to_signature(MultiLabelClassificationOutput, ModelType.dspy), ExpectedDSPy)
    assert_gliner_schemas_equal(convert_to_signature(MultiLabelClassificationOutput, ModelType.gliner, mode="classification"), expected_gliner)
    assert convert_to_signature(MultiLabelClassificationOutput, ModelType.huggingface) == expected_hf
    assert convert_to_signature(MultiLabelClassificationOutput, ModelType.outlines) == expected_outlines


def test_information_extraction_migration():
    # 1. Define Unified Pydantic Prompt Signature (for one entity)
    class Person(pydantic.BaseModel):
        """Extract person details."""
        name: str = pydantic.Field(description="Full name")
        age: int = pydantic.Field(description="Age in years")

    # 2. Define Expected Output Signatures

    # DSPy
    class ExpectedDSPy(dspy.Signature):
        """Extract person details."""
        text = dspy.InputField(desc="Input text to process.")
        name = dspy.OutputField(desc="Full name")
        age = dspy.OutputField(desc="Age in years")

    # GliNER (structure mode)
    expected_gliner = gliner2.inference.engine.Schema().structure("Person")
    expected_gliner.field("name", dtype="str")
    expected_gliner.field("age", dtype="str")

    # 3. Verify
    assert_dspy_sigs_equal(convert_to_signature(Person, ModelType.dspy), ExpectedDSPy)
    assert_gliner_schemas_equal(convert_to_signature(Person, ModelType.gliner, mode="structure"), expected_gliner)


def test_ner_migration():
    # 1. Define Unified Pydantic Prompt Signature
    class NEROutput(pydantic.BaseModel):
        """Identify named entities in the text."""
        label: Literal["PER", "LOC", "ORG"] = pydantic.Field(description="Entity type")

    # 2. Define Expected Output Signatures

    # GliNER (entities mode)
    expected_gliner = gliner2.inference.engine.Schema().entities(
        entity_types=["PER", "LOC", "ORG"]
    )

    # 3. Verify
    assert_gliner_schemas_equal(convert_to_signature(NEROutput, ModelType.gliner, mode="entities"), expected_gliner)


def test_pii_masking_migration():
    # 1. Define Unified Pydantic Prompt Signature
    class PIIOutput(pydantic.BaseModel):
        """Mask PII entities in the text and list them."""
        masked_text: str = pydantic.Field(description="Text with PII replaced by placeholders")
        pii_entities: list[str] = pydantic.Field(description="List of detected PII entities")

    # 2. Define Expected Output Signatures

    # DSPy
    class ExpectedDSPy(dspy.Signature):
        """Mask PII entities in the text and list them."""
        text = dspy.InputField(desc="Input text to process.")
        masked_text = dspy.OutputField(desc="Text with PII replaced by placeholders")
        pii_entities = dspy.OutputField(desc="List of detected PII entities")

    # 3. Verify
    assert_dspy_sigs_equal(convert_to_signature(PIIOutput, ModelType.dspy), ExpectedDSPy)


def test_question_answering_migration():
    # 1. Define Unified Pydantic Prompt Signature
    class QAOutput(pydantic.BaseModel):
        """Answer the question based on the text."""
        answer: str = pydantic.Field(description="The answer to the question")
        score: float = pydantic.Field(description="Confidence score")

    # 2. Define Expected Output Signatures

    # DSPy
    class ExpectedDSPy(dspy.Signature):
        """Answer the question based on the text."""
        text = dspy.InputField(desc="Input text to process.")
        answer = dspy.OutputField(desc="The answer to the question")
        score = dspy.OutputField(desc="Confidence score")

    # 3. Verify
    assert_dspy_sigs_equal(convert_to_signature(QAOutput, ModelType.dspy), ExpectedDSPy)


def test_relation_extraction_migration():
    # 1. Define Unified Pydantic Prompt Signature (for one triplet)
    class RelationOutput(pydantic.BaseModel):
        """Extract relations between entities."""
        head: str = pydantic.Field(description="Subject entity")
        relation: Literal["WORKS_AT", "LIVES_IN"] = pydantic.Field(description="Relation type")
        tail: str = pydantic.Field(description="Object entity")

    # 2. Define Expected Output Signatures

    # GliNER (relations mode)
    expected_gliner = gliner2.inference.engine.Schema().relations(
        relation_types=["head", "relation", "tail"] # Per instruction: all but score
    )

    # 3. Verify
    assert_gliner_schemas_equal(convert_to_signature(RelationOutput, ModelType.gliner, mode="relations"), expected_gliner)


def test_sentiment_analysis_migration():
    # 1. Define Unified Pydantic Prompt Signature
    class SentimentOutput(pydantic.BaseModel):
        """Analyze sentiment per aspect."""
        sentiment_per_aspect: dict[str, float] = pydantic.Field(description="Score per aspect")
        score: float = pydantic.Field(description="Overall confidence")

    # 2. Define Expected Output Signatures

    # DSPy
    class ExpectedDSPy(dspy.Signature):
        """Analyze sentiment per aspect."""
        text = dspy.InputField(desc="Input text to process.")
        sentiment_per_aspect = dspy.OutputField(desc="Score per aspect")
        score = dspy.OutputField(desc="Overall confidence")

    # 3. Verify
    assert_dspy_sigs_equal(convert_to_signature(SentimentOutput, ModelType.dspy), ExpectedDSPy)


def test_summarization_migration():
    # 1. Define Unified Pydantic Prompt Signature
    class SummarizationOutput(pydantic.BaseModel):
        """Summarize the text."""
        summary: str = pydantic.Field(description="The generated summary")
        score: float = pydantic.Field(description="Confidence score")

    # 2. Define Expected Output Signatures

    # DSPy
    class ExpectedDSPy(dspy.Signature):
        """Summarize the text."""
        text = dspy.InputField(desc="Input text to process.")
        summary = dspy.OutputField(desc="The generated summary")
        score = dspy.OutputField(desc="Confidence score")

    # 3. Verify
    assert_dspy_sigs_equal(convert_to_signature(SummarizationOutput, ModelType.dspy), ExpectedDSPy)


def test_translation_migration():
    # 1. Define Unified Pydantic Prompt Signature
    class TranslationOutput(pydantic.BaseModel):
        """Translate the text."""
        translation: str = pydantic.Field(description="The translated text")
        score: float = pydantic.Field(description="Confidence score")

    # 2. Define Expected Output Signatures

    # DSPy
    class ExpectedDSPy(dspy.Signature):
        """Translate the text."""
        text = dspy.InputField(desc="Input text to process.")
        translation = dspy.OutputField(desc="The translated text")
        score = dspy.OutputField(desc="Confidence score")

    # 3. Verify
    assert_dspy_sigs_equal(convert_to_signature(TranslationOutput, ModelType.dspy), ExpectedDSPy)



"""
Test file containing examples for the Custom Tasks guide.

These code blocks are referenced in docs/guides/custom_tasks.md using snippet injection.
"""

from __future__ import annotations

from typing import Sequence

import pytest


def test_basic_custom_task():
    """Test basic custom task example."""
    # --8<-- [start:custom-task-basic]
    from typing import Iterable
    from sieves.tasks.core import Task
    from sieves.data import Doc

    class CharCountTask(Task):
        def _call(self, docs: Iterable[Doc]) -> Iterable[Doc]:
            """Counts characters in doc.text.

            :param docs: Documents to process.
            :return Iterable[Doc]: Processed documents.
            """
            for doc in docs:
                doc.results[self.id] = len(doc.text)
                yield doc
    # --8<-- [end:custom-task-basic]

    # Test the task
    task = CharCountTask(task_id="CharCount", include_meta=False, batch_size=-1)
    docs = [Doc(text="Hello"), Doc(text="World!")]
    results = list(task(docs))
    assert results[0].results[task.id] == 5
    assert results[1].results[task.id] == 6


def test_custom_bridge_example():
    """Demonstrates custom bridge implementation (for docs only)."""
    # --8<-- [start:custom-bridge-sentiment]
    # --8<-- [start:custom-bridge-sentiment-imports]
    from collections.abc import Iterable
    from functools import cached_property

    import pydantic

    from sieves.data import Doc
    from sieves.model_wrappers import ModelWrapperInferenceMode, outlines_
    from sieves.tasks.predictive.bridges import Bridge
    # --8<-- [end:custom-bridge-sentiment-imports]


    # --8<-- [start:custom-bridge-sentiment-schema]
    # This is how we require our response to look like - we require not just the score, but also a reasoning/justification
    # for why this model assigns this score. We also force the score to be between 0 and 1.
    class SentimentEstimate(pydantic.BaseModel):
       reasoning: str
       score: pydantic.confloat(ge=0, le=1)
    # --8<-- [end:custom-bridge-sentiment-schema]


    # This is the bridge class.
    # --8<-- [start:custom-bridge-sentiment-class-def]
    class OutlinesSentimentAnalysis(Bridge[SentimentEstimate, SentimentEstimate, outlines_.InferenceMode]):
    # --8<-- [end:custom-bridge-sentiment-class-def]
        @property
        def model_type(self) -> ModelType:
            from sieves.model_wrappers import ModelType
            return ModelType.outlines
        @property
        def model_type(self) -> ModelType:
            from sieves.model_wrappers import ModelType
            return ModelType.outlines
        # --8<-- [start:custom-bridge-sentiment-prompt]
        @property
        def _default_prompt_instructions(self) -> str:
            return (
                "Estimate the sentiment in this text as a float between 0 and 1. 0 is negative, 1 is positive. "
                "Provide your reasoning for why you estimate this score before you output the score."
            )

        @property
        def _prompt_conclusion(self) -> str | None:
            return "========\nText: {{ text }}\nOutput:"
        # --8<-- [end:custom-bridge-sentiment-prompt]

        # --8<-- [start:custom-bridge-sentiment-properties]
        @property
        def inference_mode(self) -> outlines_.InferenceMode:
            return self._model_settings.inference_mode or outlines_.InferenceMode.json

        # We return our SentimentEstimate as prompt signature.
        @cached_property
        def prompt_signature(self) -> type[pydantic.BaseModel]:
            return SentimentEstimate
        # --8<-- [end:custom-bridge-sentiment-properties]

        # --8<-- [start:custom-bridge-sentiment-integrate]
        # We copy the result score into our doc's results attribute.
        def integrate(self, results: Sequence[SentimentEstimate], docs: list[Doc]) -> list[Doc]:
            """Integrate results into Doc instances."""
            for doc, result in zip(docs, results):
                assert isinstance(result, SentimentEstimate)
                # doc.results is a dict, with the task ID being the key to store our results under for the corresponding
                # task.
                doc.results[self._task_id] = result.score
            return docs
        # --8<-- [end:custom-bridge-sentiment-integrate]

        # --8<-- [start:custom-bridge-sentiment-consolidate]
        # Consolidating multiple chunks for sentiment analysis: we compute the average score over
        # all chunks and assume this to be the sentiment score for the entire document.
        def consolidate(
            self, results: Sequence[SentimentEstimate], docs_offsets: list[tuple[int, int]]
        ) -> Sequence[SentimentEstimate]:
            """Consolidate results for document chunks into document results."""
            consolidated_results: list[SentimentEstimate] = []

            # docs_offsets contains (start, end) tuples indicating which result indices belong to which document.
            # Example: [(0, 3), (3, 5)] means doc1 uses results[0:3], doc2 uses results[3:5]
            # This mapping is necessary because long documents are split into multiple chunks for processing.
            for doc_offset in docs_offsets:
                # Accumulate reasonings and scores from all chunks of this document
                reasonings: list[str] = []
                scores = 0.

                # Process each chunk's result for this document
                for chunk_result in results[doc_offset[0] : doc_offset[1]]:
                    # Model wrappers may return None results if they encounter errors in permissive mode.
                    # Skip None results to avoid crashes while still processing valid chunks.
                    if chunk_result:
                        assert isinstance(chunk_result, SentimentEstimate)
                        reasonings.append(chunk_result.reasoning)
                        scores += chunk_result.score

                # Calculate how many chunks this document has
                num_chunks = doc_offset[1] - doc_offset[0]

                consolidated_results.append(SentimentEstimate(
                   # Average the sentiment score across all chunks of this document
                   score=scores / num_chunks,
                   # Concatenate all chunk reasonings into a single string for the document
                   # (in production, you might want more sophisticated reasoning aggregation)
                   reasoning=str(reasonings)
                ))
            return consolidated_results
        # --8<-- [end:custom-bridge-sentiment-consolidate]
    # --8<-- [end:custom-bridge-sentiment]


def test_custom_predictive_task_example():
    """Demonstrates custom predictive task wrapper (for docs only)."""
    # --8<-- [start:custom-task-predictive]
    # --8<-- [start:custom-task-predictive-imports]
    from collections.abc import Iterable
    from typing import Any

    import datasets
    import pydantic

    from sieves.data import Doc
    from sieves.model_wrappers import ModelType
    from sieves.serialization import Config
    from sieves.tasks.predictive.core import PredictiveTask
    # --8<-- [end:custom-task-predictive-imports]

    # --8<-- [start:custom-task-predictive-schema]
    # Define the output schema for sentiment estimation
    class SentimentEstimate(pydantic.BaseModel):
       reasoning: str
       score: pydantic.confloat(ge=0, le=1)
    # --8<-- [end:custom-task-predictive-schema]

    # --8<-- [start:custom-task-predictive-bridge-imports]
    # Full bridge implementation (self-contained)
    from functools import cached_property
    from sieves.model_wrappers import ModelWrapperInferenceMode, outlines_
    from sieves.tasks.predictive.bridges import Bridge
    # --8<-- [end:custom-task-predictive-bridge-imports]

    # --8<-- [start:custom-task-predictive-bridge-class]
    class OutlinesSentimentAnalysis(Bridge[SentimentEstimate, SentimentEstimate, outlines_.InferenceMode]):
    # --8<-- [end:custom-task-predictive-bridge-class]
        @property
        def model_type(self) -> ModelType:
            from sieves.model_wrappers import ModelType
            return ModelType.outlines
        # --8<-- [start:custom-task-predictive-bridge-prompt]
        @property
        def _default_prompt_instructions(self) -> str:
            return (
                "Estimate the sentiment in this text as a float between 0 and 1. 0 is negative, 1 is positive. "
                "Provide your reasoning for why you estimate this score before you output the score."
            )

        @property
        def _prompt_conclusion(self) -> str | None:
            return "========\nText: {{ text }}\nOutput:"
        # --8<-- [end:custom-task-predictive-bridge-prompt]

        # --8<-- [start:custom-task-predictive-bridge-properties]
        @property
        def _prompt_example_template(self) -> str | None:
            return None

        @property
        def inference_mode(self) -> outlines_.InferenceMode:
            return self._model_settings.inference_mode or outlines_.InferenceMode.json

        # We return our SentimentEstimate as prompt signature.
        @cached_property
        def prompt_signature(self) -> type[pydantic.BaseModel]:
            return SentimentEstimate
        # --8<-- [end:custom-task-predictive-bridge-properties]

        # --8<-- [start:custom-task-predictive-bridge-methods]
        # We copy the result score into our doc's results attribute.
        def integrate(self, results: Sequence[SentimentEstimate], docs: list[Doc]) -> list[Doc]:
            """Integrate results into Doc instances."""
            for doc, result in zip(docs, results):
                assert isinstance(result, SentimentEstimate)
                # doc.results is a dict, with the task ID being the key to store our results under for the corresponding
                # task.
                doc.results[self._task_id] = result.score
            return docs

        # Consolidating multiple chunks for sentiment analysis can be pretty straightforward: we compute the average over
        # all chunks and assume this to be the sentiment score for the doc.
        def consolidate(
            self, results: Sequence[SentimentEstimate], docs_offsets: list[tuple[int, int]]
        ) -> Sequence[SentimentEstimate]:
            """Consolidate results for document chunks into document results."""
            consolidated_results: list[SentimentEstimate] = []

            # Iterate over indices that determine which chunks belong to which documents.
            for doc_offset in docs_offsets:
                # Keep track of all reasonings and the total score.
                reasonings: list[str] = []
                scores = 0.

                # Iterate over chunks' results.
                for chunk_result in results[doc_offset[0] : doc_offset[1]]:
                    # Model wrappers may return None results if they encounter errors and run in permissive mode. We ignore such
                    # results.
                    if chunk_result:
                        assert isinstance(chunk_result, SentimentEstimate)
                        reasonings.append(chunk_result.reasoning)
                        scores += chunk_result.score

                consolidated_results.append(SentimentEstimate(
                   # Average the score.
                   score=scores / (doc_offset[1] - doc_offset[0]),
                   # Concatenate all reasonings.
                   reasoning=str(reasonings)
                ))
            return consolidated_results
        # --8<-- [end:custom-task-predictive-bridge-methods]

    # --8<-- [start:custom-task-predictive-fewshot]
    # We'll define that class we require fewshot examples to be provided in. In our case we can just inherit from our
    # prompt signature class and add a `text` property.
    class FewshotExample(SentimentEstimate):
        text: str
    # --8<-- [end:custom-task-predictive-fewshot]


    # --8<-- [start:custom-task-predictive-task-class]
    class SentimentAnalysis(PredictiveTask[SentimentEstimate, SentimentEstimate, OutlinesSentimentAnalysis]):
    # --8<-- [end:custom-task-predictive-task-class]
        @property
        def metric(self) -> str:
            return "MSE"

        @property
        def prompt_signature(self) -> type[pydantic.BaseModel]:
            return SentimentEstimate
        # --8<-- [start:custom-task-predictive-init-supports]
        # For the initialization of the bridge. We raise an error if an model wrapper has been specified that we don't
        # support (due to us not having a bridge implemented that would support this model type).
        def _init_bridge(self, model_type: ModelType) -> OutlinesSentimentAnalysis:
            if model_type == ModelType.outlines:
                return OutlinesSentimentAnalysis(
                    task_id=self._task_id,
                    prompt_instructions=self._custom_prompt_instructions,
                    overwrite=False,
                    model_settings=self._model_settings,
                    prompt_signature=self.prompt_signature,
                )
            else:
                raise KeyError(f"Model type {model_type} is not supported by {self.__class__.__name__}.")

        # Represents set of supported model types.
        @property
        def supports(self) -> set[ModelType]:
            return {ModelType.outlines}
        # --8<-- [end:custom-task-predictive-init-supports]

        # --8<-- [start:custom-task-predictive-to-hf-dataset]
        # This implements the conversion of a set of docs to a Hugging Face datasets.Dataset.
        # You can implement this as `raise NotImplementedError` if you're not interested in generating a Hugging Face
        # dataset from your result data.
        def to_hf_dataset(self, docs: Iterable[Doc]) -> datasets.Dataset:
            # Define metadata.
            info = datasets.DatasetInfo(
                description=f"Sentiment estimation dataset. Generated with sieves"
                            f"v{Config.get_version()}.",
                features=datasets.Features({"text": datasets.Value("string"), "score": datasets.Value("float32")}),
            )

            def generate_data() -> Iterable[dict[str, Any]]:
                """Yields results as dicts.
                :return Iterable[dict[str, Any]]: Results as dicts.
                """
                for doc in docs:
                    yield {"text": doc.text, "score": doc.results[self._task_id]}

            # Create dataset.
            return datasets.Dataset.from_generator(generate_data, features=info.features, info=info)
        # --8<-- [end:custom-task-predictive-to-hf-dataset]

        # Distillation not implemented for this example task (not shown in docs)
        def distill(self, base_model_id, framework, data, output_path, val_frac, init_kwargs=None, train_kwargs=None, seed=None):
            raise NotImplementedError("Distillation not implemented for this example task")
    # --8<-- [end:custom-task-predictive]


def test_using_custom_task_example(small_outlines_model):
    """Demonstrates using the custom sentiment analysis task."""
    # First define the classes (from previous tests)
    from collections.abc import Iterable
    from functools import cached_property
    from typing import Any
    import pydantic
    import datasets
    from sieves.data import Doc
    from sieves.model_wrappers import ModelType, ModelWrapperInferenceMode, outlines_
    from sieves.tasks.predictive.bridges import Bridge
    from sieves.tasks.predictive.core import PredictiveTask
    from sieves.serialization import Config

    class SentimentEstimate(pydantic.BaseModel):
       reasoning: str
       score: pydantic.confloat(ge=0, le=1)

    class OutlinesSentimentAnalysis(Bridge[SentimentEstimate, SentimentEstimate, outlines_.InferenceMode]):
        @property
        def model_type(self) -> ModelType:
            return ModelType.outlines

        @property
        def _default_prompt_instructions(self) -> str:
            return (
                "Estimate the sentiment in this text as a float between 0 and 1. 0 is negative, 1 is positive. "
                "Provide your reasoning for why you estimate this score before you output the score."
            )

        @property
        def _prompt_conclusion(self) -> str | None:
            return "========\nText: {{ text }}\nOutput:"

        @property
        def _prompt_example_template(self) -> str | None:
            return None

        @property
        def inference_mode(self) -> outlines_.InferenceMode:
            return self._model_settings.inference_mode or outlines_.InferenceMode.json

        @cached_property
        def prompt_signature(self) -> type[pydantic.BaseModel]:
            """Return the prompt signature."""
            return SentimentEstimate

        def integrate(self, results: Sequence[SentimentEstimate], docs: list[Doc]) -> list[Doc]:
            """Integrate results into Doc instances."""
            for doc, result in zip(docs, results):
                if result:
                    assert isinstance(result, SentimentEstimate)
                    doc.results[self._task_id] = result.score
            return docs

        def consolidate(
            self, results: Sequence[SentimentEstimate], docs_offsets: list[tuple[int, int]]
        ) -> Sequence[SentimentEstimate]:
            """Consolidate results for document chunks into document results."""
            results = list(results)
            consolidated_results: list[SentimentEstimate] = []
            for doc_offset in docs_offsets:
                reasonings: list[str] = []
                scores = 0.
                for chunk_result in results[doc_offset[0] : doc_offset[1]]:
                    if chunk_result:
                        assert isinstance(chunk_result, SentimentEstimate)
                        reasonings.append(chunk_result.reasoning)
                        scores += chunk_result.score
                consolidated_results.append(SentimentEstimate(
                   score=scores / (doc_offset[1] - doc_offset[0]),
                   reasoning=str(reasonings)
                ))
            return consolidated_results

    class FewshotExample(SentimentEstimate):
        """Few-shot example for sentiment analysis."""
        text: str

    class SentimentAnalysis(PredictiveTask[SentimentEstimate, SentimentEstimate, OutlinesSentimentAnalysis]):
        """Custom sentiment analysis task."""
        @property
        def metric(self) -> str:
            return "MSE"

        @property
        def fewshot_example_type(self) -> type[FewshotExample]:
            return FewshotExample

        @property
        def prompt_signature(self) -> type[pydantic.BaseModel]:
            return SentimentEstimate

        def __init__(self, model, task_id: str = "SentimentAnalysis", include_meta: bool = True, batch_size: int = -1,
                     prompt_instructions: str | None = None, fewshot_examples: Any = (),
                     model_settings=None):
            if model_settings is None:
                from sieves.model_wrappers.types import ModelSettings
                model_settings = ModelSettings()
            super().__init__(
                model=model, task_id=task_id, include_meta=include_meta, batch_size=batch_size,
                overwrite=False, prompt_instructions=prompt_instructions, fewshot_examples=fewshot_examples,
                model_settings=model_settings, condition=None
            )

        def _init_bridge(self, model_type: ModelType) -> OutlinesSentimentAnalysis:
            if model_type == ModelType.outlines:
                return OutlinesSentimentAnalysis(
                    task_id=self._task_id,
                    prompt_instructions=self._custom_prompt_instructions,
                    overwrite=False,
                    model_settings=self._model_settings,
                    prompt_signature=self.prompt_signature,
                    model_type=model_type
                )
            else:
                raise KeyError(f"Model type {model_type} is not supported by {self.__class__.__name__}.")

        @property
        def supports(self) -> set[ModelType]:
            return {ModelType.outlines}

        def to_hf_dataset(self, docs: Iterable[Doc]) -> datasets.Dataset:
            info = datasets.DatasetInfo(
                description=f"Sentiment estimation dataset. Generated with sieves v{Config.get_version()}.",
                features=datasets.Features({"text": datasets.Value("string"), "score": datasets.Value("float32")}),
            )
            def generate_data() -> Iterable[dict[str, Any]]:
                for doc in docs:
                    yield {"text": doc.text, "score": doc.results[self._task_id]}
            return datasets.Dataset.from_generator(generate_data, features=info.features, info=info)

        def distill(self, base_model_id, framework, data, output_path, val_frac, init_kwargs=None, train_kwargs=None, seed=None):
            raise NotImplementedError("Distillation not implemented for this example task")

    # --8<-- [start:custom-task-usage]
    from sieves import Doc, Pipeline
    import outlines

    model = small_outlines_model

    docs = [Doc(text="I'm feeling happy today."), Doc(text="I was sad yesterday.")]
    pipe = Pipeline([SentimentAnalysis(model=model)])

    for doc in pipe(docs):
        print(doc.text, doc.results["SentimentAnalysis"])
    # --8<-- [end:custom-task-usage]



"""Text summarization predictive task."""

from __future__ import annotations

from collections.abc import Callable, Iterable, Sequence
from pathlib import Path
from typing import Any, override

import datasets
import dspy
import pydantic

from sieves.data import Doc
from sieves.model_wrappers import ModelType
from sieves.model_wrappers.types import ModelSettings
from sieves.serialization import Config
from sieves.tasks.distillation.types import DistillationFramework
from sieves.tasks.predictive.core import PredictiveTask
from sieves.tasks.predictive.schemas.summarization import (
    FewshotExample,
    TaskModel,
    TaskPromptSignature,
    TaskResult,
)
from sieves.tasks.predictive.summarization.bridges import (
    DSPySummarization,
    PydanticSummarization,
)

_TaskBridge = DSPySummarization | PydanticSummarization


class Summarization(PredictiveTask[TaskPromptSignature, TaskResult, _TaskBridge]):
    """Summarize documents to a target length using structured model wrappers."""

    def __init__(
        self,
        n_words: int,
        model: TaskModel,
        task_id: str | None = None,
        include_meta: bool = True,
        batch_size: int = -1,
        overwrite: bool = False,
        prompt_instructions: str | None = None,
        fewshot_examples: Sequence[FewshotExample] = (),
        model_settings: ModelSettings = ModelSettings(),
        condition: Callable[[Doc], bool] | None = None,
    ) -> None:
        """Initialize new Summarization task.

        :param n_words: Maximal number of words (consider this a guideline, not a strict limit).
        :param model: Model to use.
        :param task_id: Task ID.
        :param include_meta: Whether to include meta information generated by the task.
        :param batch_size: Batch size to use for inference. Use -1 to process all documents at once.
        :param overwrite: Some tasks, e.g. anonymization or translation, output a modified version of the input text.
            If True, these tasks overwrite the original document text. If False, the result will just be stored in the
            documents' `.results` field.
        :param prompt_instructions: Custom prompt instructions. If None, default instructions are used.
        :param fewshot_examples: Few-shot examples.
        :param model_settings: Settings for structured generation. Use the `inference_mode` field to specify the
            inference mode for the model wrapper.
        :param condition: Optional callable that determines whether to process each document.
        """
        self._n_words = n_words

        super().__init__(
            model=model,
            task_id=task_id,
            include_meta=include_meta,
            batch_size=batch_size,
            overwrite=overwrite,
            prompt_instructions=prompt_instructions,
            fewshot_examples=fewshot_examples,
            model_settings=model_settings,
            condition=condition,
        )

    @property
    @override
    def fewshot_example_type(self) -> type[FewshotExample]:
        """Return few-shot example type.

        :return: Few-shot example type.
        """
        return FewshotExample

    @property
    @override
    def prompt_signature(self) -> type[pydantic.BaseModel]:
        # Dynamically create type with length restriction in description.
        class ResultWithLengthRestriction(TaskResult):
            pass

        ResultWithLengthRestriction.model_fields[
            "summary"
        ].description += f" This summary should be around {self._n_words} words."

        return ResultWithLengthRestriction

    @override
    def _compute_metrics(self, truths: list[Any], preds: list[Any], judge: dspy.LM | None = None) -> dict[str, float]:
        """Compute corpus-level metrics.

        :param truths: List of ground truths.
        :param preds: List of predictions.
        :param judge: Optional DSPy LM instance to use as judge for generative tasks.
        :return: Dictionary of metrics.
        """
        for gold in truths:
            if gold is not None:
                assert isinstance(gold, TaskResult)
        for pred in preds:
            if pred is not None:
                assert isinstance(pred, TaskResult)

        return super()._compute_metrics(truths, preds, judge=judge)

    @override
    def _init_bridge(self, model_type: ModelType) -> _TaskBridge:
        bridge_types: dict[ModelType, type[DSPySummarization | PydanticSummarization]] = {
            ModelType.dspy: DSPySummarization,
            ModelType.langchain: PydanticSummarization,
            ModelType.outlines: PydanticSummarization,
        }

        try:
            return bridge_types[model_type](
                task_id=self._task_id,
                prompt_instructions=self._custom_prompt_instructions,
                n_words=self._n_words,
                overwrite=self._overwrite,
                model_settings=self._model_settings,
                prompt_signature=self.prompt_signature,
                model_type=model_type,
                fewshot_examples=self._fewshot_examples,
            )
        except KeyError as err:
            raise KeyError(f"Model type {model_type} is not supported by {self.__class__.__name__}.") from err

    @staticmethod
    @override
    def supports() -> set[ModelType]:
        return {
            ModelType.dspy,
            ModelType.langchain,
            ModelType.outlines,
        }

    @property
    @override
    def _state(self) -> dict[str, Any]:
        return {
            **super()._state,
            "n_words": self._n_words,
        }

    @override
    def to_hf_dataset(self, docs: Iterable[Doc], threshold: float | None = None) -> datasets.Dataset:
        # Define metadata.
        features = datasets.Features(
            {"text": datasets.Value("string"), "summary": datasets.Value("string"), "score": datasets.Value("float32")}
        )
        info = datasets.DatasetInfo(
            description=f"Summarization dataset. Generated with sieves v{Config.get_version()}.",
            features=features,
        )

        # Fetch data used for generating dataset.
        try:
            data = [(doc.text, doc.results[self._task_id].summary, doc.results[self._task_id].score) for doc in docs]
        except KeyError as err:
            raise KeyError(f"Not all documents have results for this task with ID {self._task_id}") from err

        def generate_data() -> Iterable[dict[str, Any]]:
            """Yield results as dicts.

            :return: Results as dicts.
            """
            for text, summary, score in data:
                yield {"text": text, "summary": summary, "score": score}

        # Create dataset.
        return datasets.Dataset.from_generator(generate_data, features=features, info=info)

    @override
    def distill(
        self,
        base_model_id: str,
        framework: DistillationFramework,
        data: datasets.Dataset | Sequence[Doc],
        output_path: Path | str,
        val_frac: float,
        init_kwargs: dict[str, Any] | None = None,
        train_kwargs: dict[str, Any] | None = None,
        seed: int | None = None,
    ) -> None:
        raise NotImplementedError



"""Bridges for summarization task."""

import abc
from collections.abc import Callable, Sequence
from typing import Any, TypeVar, override

import dspy
import pydantic

from sieves.data import Doc
from sieves.model_wrappers import (
    ModelType,
    ModelWrapperInferenceMode,
    dspy_,
    langchain_,
    outlines_,
)
from sieves.model_wrappers.types import ModelSettings
from sieves.tasks.predictive.bridges import Bridge
from sieves.tasks.predictive.consolidation import TextConsolidation
from sieves.tasks.predictive.schemas.summarization import Result

_BridgePromptSignature = TypeVar("_BridgePromptSignature")
_BridgeResult = TypeVar("_BridgeResult")


class SummarizationBridge(Bridge[_BridgePromptSignature, _BridgeResult, ModelWrapperInferenceMode], abc.ABC):
    """Abstract base class for summarization bridges."""

    def __init__(
        self,
        task_id: str,
        prompt_instructions: str | None,
        n_words: int,
        overwrite: bool,
        model_settings: ModelSettings,
        prompt_signature: type[pydantic.BaseModel],
        model_type: ModelType,
        fewshot_examples: Sequence[pydantic.BaseModel] = (),
    ):
        """Initialize summarization bridge.

        :param task_id: Task ID.
        :param prompt_instructions: Custom prompt instructions. If None, default instructions are used.
        :param n_words: Maximum number of words for summary.
        :param overwrite: Whether to overwrite original text with summary.
        :param model_settings: Settings for structured generation.
        :param prompt_signature: Unified Pydantic prompt signature.
        :param model_type: Model type.
        :param fewshot_examples: Few-shot examples.
        """
        super().__init__(
            task_id=task_id,
            prompt_instructions=prompt_instructions,
            overwrite=overwrite,
            model_settings=model_settings,
            prompt_signature=prompt_signature,
            model_type=model_type,
            fewshot_examples=fewshot_examples,
        )
        self._n_words = n_words
        self._consolidation_strategy = TextConsolidation(extractor=self._chunk_extractor)

    @property
    @abc.abstractmethod
    def _chunk_extractor(self) -> Callable[[Any], tuple[str, float | None]]:
        """Return a callable that extracts (text, score) from a raw chunk result.

        :return: Extractor callable.
        """


class DSPySummarization(SummarizationBridge[dspy_.PromptSignature, dspy_.Result, dspy_.InferenceMode]):
    """DSPy bridge for summarization."""

    @override
    def _validate(self) -> None:
        assert self._model_type == ModelType.dspy

    @override
    @property
    def _default_prompt_instructions(self) -> str:
        return ""

    @override
    @property
    def inference_mode(self) -> dspy_.InferenceMode:
        return self._model_settings.inference_mode or dspy_.InferenceMode.predict

    @property
    @override
    def _chunk_extractor(self) -> Callable[[Any], tuple[str, float | None]]:
        return lambda res: (res.summary, getattr(res, "score", None))

    @override
    def integrate(self, results: Sequence[dspy_.Result], docs: list[Doc]) -> list[Doc]:
        for doc, result in zip(docs, results):
            assert len(result.completions.summary) == 1
            res = Result(summary=result.summary, score=getattr(result, "score", None))
            doc.results[self._task_id] = res

            if self._overwrite:
                doc.text = res.summary

        return docs

    @override
    def consolidate(
        self, results: Sequence[dspy_.Result], docs_offsets: list[tuple[int, int]]
    ) -> Sequence[dspy_.Result]:
        consolidated_results_clean = self._consolidation_strategy.consolidate(results, docs_offsets)

        # Wrap back into dspy.Prediction.
        consolidated_results: list[dspy_.Result] = []
        for summary, score in consolidated_results_clean:
            consolidated_results.append(
                dspy.Prediction.from_completions(
                    {
                        "summary": [summary],
                        "score": [score],
                    },
                    signature=self.prompt_signature,
                )
            )

        return consolidated_results


class PydanticSummarization(SummarizationBridge[pydantic.BaseModel, pydantic.BaseModel, ModelWrapperInferenceMode]):
    """Pydantic-based summarization bridge."""

    @override
    def _validate(self) -> None:
        assert self._model_type in {ModelType.langchain, ModelType.outlines}

    @override
    @property
    def _default_prompt_instructions(self) -> str:
        return (
            "Your goal is to summarize a text. Provide a confidence score between 0.0 and 1.0 for the summary. "
            f"The number of words should be around {self._n_words}."
        )

    @override
    @property
    def _prompt_conclusion(self) -> str:
        return "========\n<text>{{ text }}</text>\n<summary>"

    @property
    @override
    def _chunk_extractor(self) -> Callable[[Any], tuple[str, float | None]]:
        return lambda res: (res.summary, getattr(res, "score", None))

    @override
    def integrate(self, results: Sequence[pydantic.BaseModel], docs: list[Doc]) -> list[Doc]:
        for doc, result in zip(docs, results):
            assert hasattr(result, "summary")
            res = Result(summary=result.summary, score=getattr(result, "score", None))
            doc.results[self._task_id] = res

            if self._overwrite:
                doc.text = res.summary
        return docs

    @override
    def consolidate(
        self, results: Sequence[pydantic.BaseModel], docs_offsets: list[tuple[int, int]]
    ) -> Sequence[pydantic.BaseModel]:
        assert issubclass(self.prompt_signature, pydantic.BaseModel)

        consolidated_results_clean = self._consolidation_strategy.consolidate(results, docs_offsets)
        consolidated_results: list[pydantic.BaseModel] = []

        for summary, score in consolidated_results_clean:
            consolidated_results.append(
                self.prompt_signature(
                    summary=summary,
                    score=score,
                )
            )

        return consolidated_results

    @override
    @property
    def inference_mode(self) -> outlines_.InferenceMode | langchain_.InferenceMode:
        if self._model_type == ModelType.outlines:
            return self._model_settings.inference_mode or outlines_.InferenceMode.json
        elif self._model_type == ModelType.langchain:
            return self._model_settings.inference_mode or langchain_.InferenceMode.structured

        raise ValueError(f"Unsupported model type: {self._model_type}")



import marimo

__generated_with = "0.18.3"
app = marimo.App(width="full")


@app.cell
def _():
    import marimo as mo

    return (mo,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # OSA Community Demo: Sieves

    2025-12-12
    """)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## NLP 101, contemporary: building a spam filter
    """)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    A classic first example in NLP is building a spam filter, often done using bag-of-words/tf-idf classification.

    How does a zero-shot version of this look like?

    We'll first define a bunch of documents. Half of them is spam, the other half is not.
    """)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    Building a dataset
    """)
    return


@app.cell
def _():
    from sieves import Doc

    docs = [
        Doc(
            text="""
        Dear User,
        We detected unusual activity on your account. To avoid immediate suspension, please verify your information by clicking the secure link below within 24 hours. Failure to comply may result in permanent closure.
        """
        ),
        Doc(
            text="""
        Hi Team,
        I wanted to share a quick update on the project. The initial phase is complete, and we’re on track to begin the next steps early next week. I’ll follow up with a detailed timeline shortly.
        """
        ),
        Doc(
            text="""
        Hello,
        You are the lucky winner of an exclusive reward valued at $5,000. This offer is time-sensitive. Confirm your eligibility now to claim your prize before it expires.
        """
        ),
        Doc(
            text="""
        Hello,
        Thank you for the productive discussion earlier today. As agreed, I’ll send over the revised proposal by Friday. Please let me know if you have any questions in the meantime.
        """
        ),
        Doc(
            text="""
        Dear Customer,
        Your recent invoice remains unpaid. Please review the attached document and settle the balance immediately to avoid additional fees. Contact support if you believe this is an error.
        """
        ),
        Doc(
            text="""
        Hi,
        Could you please review the attached document and share your feedback by the end of the week? Your input will help us finalize the draft before submission.
        """
        ),
    ]
    return (docs,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    First, we'll define our model. In this case we go with DSPy (which uses LiteLLM underneath), but other libraries
    are supported:
    - Hugging Face zero-shot pipelines
    - GliNER2
    - Outlines
    - LangChain
    """)
    return


@app.cell
def _():
    import os

    import anthropic
    import dspy
    import outlines
    import transformers

    openrouter_api_base = "https://openrouter.ai/api/v1/"
    openrouter_model_id = "google/gemini-2.5-flash-lite-preview-09-2025"

    model = dspy.LM(
        f"openrouter/{openrouter_model_id}", api_base=openrouter_api_base, api_key=os.environ["OPENROUTER_API_KEY"]
    )
    return (model,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## Building a pipeline

    We'll want to **classify** our data into spam and not spam. We'd also like a **summary** of our mails.
    """)
    return


@app.cell
def _(model):
    from sieves import tasks

    classifier = tasks.Classification(
        labels=["spam", "not spam"],
        mode='single',
        model=model,
    )
    summarizer = tasks.Summarization(n_words=10, model=model)
    pipe = classifier + summarizer
    return classifier, pipe, tasks


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## Spam or not spam?

    Let's give this a spin:
    """)
    return


@app.cell
def _(docs, pipe):
    from pprint import pprint

    for doc in pipe(docs):
        print(doc.text)
        pprint(doc.results["Classification"])
        pprint(doc.results["Summarization"])
    return (pprint,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    Looks reasonable! But wait, we're not interested in summarizing spam emails - we just want to ignore them.

    ## Enter conditionals
    """)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    Conditionals are functions that tell the pipeline which documents to skip for a task. For example, we'll want to skip summarizing spam emails, as this is wasted time and money:
    """)
    return


@app.cell
def _(classifier, docs, model, pprint, tasks):
    conditional_summarizer = tasks.Summarization(
        n_words=10, model=model, condition=lambda doc: doc.results["Classification"][0] == "not spam"
    )
    improved_pipe = classifier + conditional_summarizer

    for doc2 in improved_pipe(docs):
        print(doc2.text)
        pprint(doc2.results["Classification"])
        pprint(doc2.results["Summarization"])
    return


@app.cell
def _():
    return


if __name__ == "__main__":
    app.run()



"""
Tests for preprocessing task usage documentation snippets.
"""
import pytest
from sieves import tasks

def test_docling_usage():
    pytest.importorskip("docling")
    # --8<-- [start:docling-usage]
    from sieves.tasks.preprocessing.ingestion.docling_ import Docling

    # Basic usage
    task = Docling(export_format="markdown")
    # --8<-- [end:docling-usage]
    assert task

def test_chonkie_usage():
    try:
        import tokenizers
        tokenizers.Tokenizer.from_pretrained("gpt2")
    except Exception:
        pytest.skip("tokenizers not ready")

    # --8<-- [start:chonkie-usage]
    import chonkie
    import tokenizers
    from sieves.tasks.preprocessing.chunking.chonkie_ import Chonkie

    # Setup tokenizer and chunker
    tokenizer = tokenizers.Tokenizer.from_pretrained("gpt2")
    chunker = chonkie.TokenChunker(tokenizer)

    task = Chonkie(chunker=chunker)
    # --8<-- [end:chonkie-usage]
    assert task

def test_naive_usage():
    # --8<-- [start:naive-usage]
    from sieves.tasks.preprocessing.chunking.naive import NaiveChunker

    # Chunk by sentence count interval
    task = NaiveChunker(interval=5)
    # --8<-- [end:naive-usage]
    assert task



"""Bridges for classification task."""

import abc
from collections.abc import Callable, Sequence
from typing import Any, Literal, TypeVar, override

import dspy
import pydantic

from sieves.data import Doc
from sieves.model_wrappers import (
    ModelType,
    ModelWrapperInferenceMode,
    dspy_,
    huggingface_,
    langchain_,
    outlines_,
)
from sieves.model_wrappers.types import ModelSettings
from sieves.tasks.predictive.bridges import Bridge
from sieves.tasks.predictive.consolidation import LabelScoreConsolidation
from sieves.tasks.predictive.schemas.classification import ResultMultiLabel, ResultSingleLabel
from sieves.tasks.predictive.utils import convert_to_signature

_BridgePromptSignature = TypeVar("_BridgePromptSignature")
_BridgeResult = TypeVar("_BridgeResult")


class ClassificationBridge(Bridge[_BridgePromptSignature, _BridgeResult, ModelWrapperInferenceMode], abc.ABC):
    """Abstract base class for classification bridges."""

    def __init__(
        self,
        task_id: str,
        prompt_instructions: str | None,
        labels: list[str] | dict[str, str],
        mode: Literal["single", "multi"],
        model_settings: ModelSettings,
        prompt_signature: type[pydantic.BaseModel],
        model_type: ModelType,
        fewshot_examples: Sequence[pydantic.BaseModel] = (),
    ):
        """Initialize ClassificationBridge.

        :param task_id: Task ID.
        :param prompt_instructions: Custom prompt instructions. If None, default instructions are used.
        :param labels: Labels to classify. Can be a list of label strings, or a dict mapping labels to descriptions.
        :param mode: If 'multi'', task returns scores for all specified labels. If 'single', task returns
        most likely class label.
        :param model_settings: Model settings.
        :param prompt_signature: Unified Pydantic prompt signature.
        :param model_type: Model type.
        :param fewshot_examples: Few-shot examples.
        """
        super().__init__(
            task_id=task_id,
            prompt_instructions=prompt_instructions,
            overwrite=False,
            model_settings=model_settings,
            prompt_signature=prompt_signature,
            model_type=model_type,
            fewshot_examples=fewshot_examples,
        )
        if isinstance(labels, dict):
            self._labels = list(labels.keys())
            self._label_descriptions = labels
        else:
            self._labels = labels
            self._label_descriptions = {}

        self._mode = mode
        self._consolidation_strategy = LabelScoreConsolidation(
            labels=self._labels,
            mode=self._mode,
            extractor=self._chunk_extractor,
        )

    @override
    @property
    def prompt_signature(self) -> _BridgePromptSignature:
        return convert_to_signature(
            model_cls=self._pydantic_signature,
            model_type=self.model_type,
            mode="classification",
        )  # type: ignore[return-value]

    @property
    @abc.abstractmethod
    def _chunk_extractor(self) -> Callable[[Any], dict[str, float]]:
        """Return a callable that extracts label scores from a raw chunk result.

        :return: Extractor callable.
        """

    def _get_label_descriptions(self) -> str:
        """Return a string with the label descriptions.

        :return: A string with the label descriptions.
        """
        labels_with_descriptions: list[str] = []
        for label in self._labels:
            if label in self._label_descriptions:
                labels_with_descriptions.append(
                    f"  <label_description>\n    <label>{label}</label>\n    <description>"
                    f"{self._label_descriptions[label]}</description>\n  </label_description>"
                )
            else:
                labels_with_descriptions.append(f"  <label>{label}</label>")

        label_desc_string = "\n".join(labels_with_descriptions)

        return f"<label_descriptions>\n{label_desc_string}\n</label_descriptions>"


class DSPyClassification(ClassificationBridge[dspy_.PromptSignature, dspy_.Result, dspy_.InferenceMode]):
    """DSPy bridge for classification."""

    @override
    def _validate(self) -> None:
        assert self._model_type == ModelType.dspy

    @override
    @property
    def _default_prompt_instructions(self) -> str:
        return ""

    @override
    @property
    def inference_mode(self) -> dspy_.InferenceMode:
        return self._model_settings.inference_mode or dspy_.InferenceMode.predict

    @property
    @override
    def _chunk_extractor(self) -> Callable[[Any], dict[str, float]]:
        def extractor(res: Any) -> dict[str, float]:
            if self._mode == "multi":
                return {label: getattr(res, label) for label in self._labels}
            return {res.label: res.score}

        return extractor

    @override
    def integrate(self, results: Sequence[dspy_.Result], docs: list[Doc]) -> list[Doc]:
        for doc, result in zip(docs, results):
            # The result is a dspy.Prediction where completions contain the fields.
            # We take the first completion.
            prediction = result.completions[0]

            if self._mode == "multi":
                label_scores = [(label, float(getattr(prediction, label))) for label in self._labels]
                sorted_preds = sorted(label_scores, key=lambda x: x[1], reverse=True)
                doc.results[self._task_id] = ResultMultiLabel(label_scores=sorted_preds)
            else:
                doc.results[self._task_id] = ResultSingleLabel(label=prediction.label, score=float(prediction.score))

        return docs

    @override
    def consolidate(
        self, results: Sequence[dspy_.Result], docs_offsets: list[tuple[int, int]]
    ) -> Sequence[dspy_.Result]:
        consolidated_results_clean = self._consolidation_strategy.consolidate(results, docs_offsets)

        # Wrap back into dspy.Prediction.
        consolidated_results: list[dspy_.Result] = []
        for scores_list in consolidated_results_clean:
            if self._mode == "multi":
                data = {label: score for label, score in scores_list}
            else:
                data = {"label": scores_list[0][0], "score": scores_list[0][1]}

            consolidated_results.append(
                dspy.Prediction.from_completions(
                    [data],
                    signature=self.prompt_signature,
                )
            )
        return consolidated_results


class HuggingFaceClassification(ClassificationBridge[list[str], huggingface_.Result, huggingface_.InferenceMode]):
    """HuggingFace bridge for classification."""

    @override
    def _validate(self) -> None:
        assert self._model_type == ModelType.huggingface

    @override
    @property
    def _default_prompt_instructions(self) -> str:
        return f"This text is about {{}}.\n{self._get_label_descriptions()}"

    @override
    @property
    def inference_mode(self) -> huggingface_.InferenceMode:
        return self._model_settings.inference_mode or huggingface_.InferenceMode.zeroshot_cls

    @property
    @override
    def _chunk_extractor(self) -> Callable[[Any], dict[str, float]]:
        # For HuggingFace zero-shot, prompt_signature is a list of field names.
        # convert_to_signature for HF returns all fields but 'score'.
        # The raw result 'res' from HF has 'labels' and 'scores'.
        return lambda res: dict(zip(res["labels"], res["scores"]))

    @override
    def integrate(self, results: Sequence[huggingface_.Result], docs: list[Doc]) -> list[Doc]:
        for doc, result in zip(docs, results):
            # result is a dict with 'labels' and 'scores'.
            # We map them back to the unified schema.
            # In multi-label mode, doc.results[self._task_id] expects a ResultMultiLabel (list of tuples).
            label_scores = list(zip(result["labels"], result["scores"]))
            sorted_preds = sorted(label_scores, key=lambda x: x[1], reverse=True)

            if self._mode == "multi":
                doc.results[self._task_id] = ResultMultiLabel(label_scores=sorted_preds)
            else:
                # In single mode, the top label should be in the ResultSingleLabel.
                # Usually HF zero-shot with mode='multi' (default in sieves) returns multiple labels.
                # ClassificationTask._init_bridge should ideally handle this.
                doc.results[self._task_id] = ResultSingleLabel(label=sorted_preds[0][0], score=sorted_preds[0][1])
        return docs

    @override
    def consolidate(
        self, results: Sequence[huggingface_.Result], docs_offsets: list[tuple[int, int]]
    ) -> Sequence[huggingface_.Result]:
        consolidated_results_clean = self._consolidation_strategy.consolidate(results, docs_offsets)

        consolidated_results: list[huggingface_.Result] = []
        for scores_list in consolidated_results_clean:
            consolidated_results.append(
                {
                    "labels": [label for label, _ in scores_list],
                    "scores": [score for _, score in scores_list],
                }
            )
        return consolidated_results


class LangChainClassification(
    ClassificationBridge[pydantic.BaseModel | list[str], pydantic.BaseModel | str, ModelWrapperInferenceMode], abc.ABC
):
    """Base class for Pydantic-based classification bridges."""

    @override
    def _validate(self) -> None:
        assert self._model_type == ModelType.langchain

    @override
    @property
    def _default_prompt_instructions(self) -> str:
        if self._mode == "multi":
            return """
            Perform multi-label classification of the provided text.
            For each label, provide a score between 0.0 (not applicable) and 1.0 (highly applicable).
            """

        return """
        Classify the provided text.
        Provide a score reflecting how likely it is that your chosen label is the correct
        fit for the text.
        """

    @override
    @property
    def _prompt_conclusion(self) -> str | None:
        return """
        ========

        <text>{{ text }}</text>
        """

    @property
    @override
    def _chunk_extractor(self) -> Callable[[Any], dict[str, float]]:
        def extractor(res: Any) -> dict[str, float]:
            if self._mode == "multi":
                return {label: float(getattr(res, label)) for label in self._labels}
            return {str(getattr(res, "label")): float(getattr(res, "score"))}

        return extractor

    @override
    def integrate(self, results: Sequence[pydantic.BaseModel | str], docs: list[Doc]) -> list[Doc]:
        for doc, result in zip(docs, results):
            if self._mode == "multi":
                assert isinstance(result, pydantic.BaseModel)
                label_scores = result.model_dump()
                sorted_label_scores = sorted(
                    ((label, score) for label, score in label_scores.items()), key=lambda x: x[1], reverse=True
                )
                doc.results[self._task_id] = ResultMultiLabel(label_scores=sorted_label_scores)

            else:
                assert hasattr(result, "label") and hasattr(result, "score")
                doc.results[self._task_id] = ResultSingleLabel(label=result.label, score=result.score)

        return docs

    @override
    def consolidate(
        self, results: Sequence[pydantic.BaseModel | str], docs_offsets: list[tuple[int, int]]
    ) -> Sequence[pydantic.BaseModel | str]:
        consolidated_results_clean = self._consolidation_strategy.consolidate(results, docs_offsets)

        consolidated_results: list[pydantic.BaseModel | str] = []
        prompt_signature = self.prompt_signature
        assert issubclass(prompt_signature, pydantic.BaseModel)  # type: ignore[arg-type]

        for scores_list in consolidated_results_clean:
            if self._mode == "multi":
                consolidated_results.append(prompt_signature(**dict(scores_list)))
            else:
                # In single mode, we only take the top label.
                top_label, top_score = scores_list[0]
                consolidated_results.append(
                    prompt_signature(
                        label=top_label,
                        score=top_score,
                    )
                )

        return consolidated_results

    @override
    @property
    def inference_mode(self) -> langchain_.InferenceMode:
        return self._model_settings.inference_mode or langchain_.InferenceMode.structured


class OutlinesClassification(LangChainClassification[ModelWrapperInferenceMode], abc.ABC):
    """Base class for Outlines-based classification bridges with label forcing."""

    @override
    def _validate(self) -> None:
        assert self._model_type == ModelType.outlines

    @override
    @property
    def _default_prompt_instructions(self) -> str:
        if self._mode == "multi":
            return super()._default_prompt_instructions

        return f"""
        Perform single-label classification of the provided text given the provided labels: {",".join(self._labels)}.
        {self._get_label_descriptions()}

        Provide the best-fitting label for given text.
        """

    @property
    @override
    def _chunk_extractor(self) -> Callable[[Any], dict[str, float]]:
        if self._mode == "multi":
            return super()._chunk_extractor

        def extractor(res: Any) -> dict[str, float]:
            if isinstance(res, str):
                return {res: 1.0}
            return {str(getattr(res, "label")): float(getattr(res, "score", 1.0))}

        return extractor

    @override
    def integrate(self, results: Sequence[pydantic.BaseModel | str], docs: list[Doc]) -> list[Doc]:
        if self._mode == "multi":
            return super().integrate(results, docs)

        for doc, result in zip(docs, results):
            # Outlines choice mode returns just the label string.
            if isinstance(result, str):
                doc.results[self._task_id] = ResultSingleLabel(label=result, score=1.0)
            else:
                # Fallback for other pydantic-based bridges.
                assert hasattr(result, "label")
                doc.results[self._task_id] = ResultSingleLabel(label=result.label, score=getattr(result, "score", 1.0))
        return docs

    @override
    def consolidate(
        self, results: Sequence[pydantic.BaseModel | str], docs_offsets: list[tuple[int, int]]
    ) -> Sequence[pydantic.BaseModel | str]:
        if self._mode == "multi":
            return super().consolidate(results, docs_offsets)

        consolidated_results_clean = self._consolidation_strategy.consolidate(results, docs_offsets)

        consolidated_results: list[pydantic.BaseModel | str] = []
        for scores_list in consolidated_results_clean:
            top_label, _ = scores_list[0]
            consolidated_results.append(top_label)

        return consolidated_results

    @override
    @property
    def inference_mode(self) -> outlines_.InferenceMode:
        return self._model_settings.inference_mode or (
            outlines_.InferenceMode.json if self._mode == "multi" else outlines_.InferenceMode.choice
        )



# mypy: ignore-errors
import pytest
from flaky import flaky

from sieves import Doc, Pipeline, tasks
from sieves.model_wrappers import ModelType, ModelSettings, dspy_, langchain_, outlines_
from sieves.serialization import Config
from sieves.tasks import PredictiveTask, PIIMasking
from sieves.tasks.predictive import pii_masking
from sieves.tasks.predictive.schemas.pii_masking import Result, PIIEntity


@pytest.mark.parametrize(
    "batch_runtime",
    PIIMasking.supports(),
    indirect=["batch_runtime"],
)
@pytest.mark.parametrize("fewshot", [True, False])
def test_run(pii_masking_docs, batch_runtime, fewshot) -> None:
    fewshot_examples = [
        pii_masking.FewshotExample(
            text="His name is John Doe and his SSN is 111-222-333.",
            masked_text="His name is [MASKED] and his SSN is [MASKED].",
            pii_entities=[
                pii_masking.PIIEntity(entity_type="PERSON", text="John Doe", score=1.0),
                pii_masking.PIIEntity(entity_type="SSN", text="111-222-333", score=1.0),
            ],
        ),
        pii_masking.FewshotExample(
            text="Contact Maria at maria.doe@gmail.com.",
            masked_text="Contact [MASKED] at [MASKED].",
            pii_entities=[
                pii_masking.PIIEntity(entity_type="PERSON", text="Maria", score=1.0),
                pii_masking.PIIEntity(entity_type="EMAIL", text="maria.doe@gmail.com", score=1.0),
            ],
        ),
    ]

    fewshot_args = {"fewshot_examples": fewshot_examples} if fewshot else {}
    task = tasks.predictive.PIIMasking(
        model=batch_runtime.model,
        model_settings=batch_runtime.model_settings,
        batch_size=batch_runtime.batch_size,
        **fewshot_args,
    )
    pipe = Pipeline(task)
    docs = list(pipe(pii_masking_docs))

    assert len(docs) == 2
    for doc in docs:
        assert doc.text
        assert "PIIMasking" in doc.results

        # Verify unified result types.
        res = doc.results["PIIMasking"]
        assert isinstance(res, pii_masking.Result)
        assert len(res.pii_entities)

        assert "PIIMasking" in doc.meta
        assert "raw" in doc.meta["PIIMasking"]
        assert "usage" in doc.meta["PIIMasking"]
        assert "usage" in doc.meta

        print(f"Output: {doc.results['PIIMasking']}")
        print(f"Raw output: {doc.meta['PIIMasking']['raw']}")
        print(f"Usage: {doc.meta['PIIMasking']['usage']}")
        print(f"Total Usage: {doc.meta['usage']}")

    with pytest.raises(NotImplementedError):
        pipe["PIIMasking"].distill(None, None, None, None, None, None, None, None)

    if fewshot:
        _to_hf_dataset(task, docs)


def _to_hf_dataset(task: PIIMasking, docs: list[Doc]) -> None:
    """Tests whether conversion to HF dataset works as expected.

    :param task: PIIMasking task instance.
    :param docs: List of documents to convert.
    """
    dataset = task.to_hf_dataset(docs)
    assert all([key in dataset.features for key in ("text", "masked_text")])
    assert len(dataset) == 2
    records = list(dataset)
    assert records[0]["text"] == "Her SSN is 222-333-444. Her credit card number is 1234 5678."
    assert records[1]["text"] == "You can reach Michael at michael.michaels@gmail.com."

    with pytest.raises(KeyError):
        task.to_hf_dataset([Doc(text="This is a dummy text.")])


@pytest.mark.parametrize("batch_runtime", [ModelType.dspy], indirect=["batch_runtime"])
def test_serialization(pii_masking_docs, batch_runtime) -> None:
    pipe = Pipeline([
        tasks.predictive.PIIMasking(
            model=batch_runtime.model,
            model_settings=batch_runtime.model_settings,
            batch_size=batch_runtime.batch_size,
        )
    ])

    config = pipe.serialize()
    assert config.model_dump() == {'cls_name': 'sieves.pipeline.core.Pipeline',
 'tasks': {'is_placeholder': False,
           'value': [{'cls_name': 'sieves.tasks.predictive.pii_masking.core.PIIMasking',
                      'fewshot_examples': {'is_placeholder': False,
                                           'value': ()},
                      'batch_size': {'is_placeholder': False, "value": -1},
                      'model_settings': {'is_placeholder': False,
                                              'value': {
                                                        'config_kwargs': None,
                                                        'inference_kwargs': None,
                                                        'init_kwargs': None,
                                                        'strict': True, 'inference_mode': None}},
                      'include_meta': {'is_placeholder': False, 'value': True},
                      'model': {'is_placeholder': True,
                                'value': 'dspy.clients.lm.LM'},
                      'pii_types': {'is_placeholder': False, 'value': None},
                      'prompt_instructions': {'is_placeholder': False,
                                          'value': None},
                      'task_id': {'is_placeholder': False,
                                  'value': 'PIIMasking'},
                      'condition': {'is_placeholder': False, 'value': None},
                      'version': Config.get_version()}]},
 'use_cache': {'is_placeholder': False, 'value': True},
 'version': Config.get_version()}

    Pipeline.deserialize(config=config, tasks_kwargs=[{"model": batch_runtime.model}])


@pytest.mark.parametrize(
    "batch_runtime",
    PIIMasking.supports(),
    indirect=["batch_runtime"],
)
def test_inference_mode_override(batch_runtime) -> None:
    """Test that inference_mode parameter overrides the default value."""
    dummy = "dummy_inference_mode"

    task = tasks.predictive.PIIMasking(
        model=batch_runtime.model,
        model_settings=ModelSettings(inference_mode=dummy),
        batch_size=batch_runtime.batch_size,
    )

    assert task._bridge.inference_mode == dummy


@pytest.mark.parametrize(
    "batch_runtime",
    PIIMasking.supports(),
    indirect=["batch_runtime"],
)
def test_run_with_dict_pii_types(pii_masking_docs, batch_runtime) -> None:
    """Test PIIMasking with dict format pii_types (labels with descriptions)."""
    pii_types_with_descriptions = {
        "EMAIL": "Email addresses",
        "PHONE": "Phone numbers",
        "SSN": "Social security numbers",
        "CREDIT_CARD": "Credit card numbers"
    }

    pipe = Pipeline([
        tasks.predictive.PIIMasking(
            pii_types=pii_types_with_descriptions,
            model=batch_runtime.model,
            model_settings=batch_runtime.model_settings,
            batch_size=batch_runtime.batch_size,
        )
    ])
    docs = list(pipe(pii_masking_docs))

    assert len(docs) == 2
    for doc in docs:
        assert doc.text
        assert "PIIMasking" in doc.results


@flaky(max_runs=3, min_passes=1)
@pytest.mark.parametrize(
    "batch_runtime",
    PIIMasking.supports(),
    indirect=["batch_runtime"],
)
def test_run_with_list_pii_types(pii_masking_docs, batch_runtime) -> None:
    """Test PIIMasking with list format pii_types (backward compatibility)."""
    pii_types_list = ["EMAIL", "PHONE", "SSN", "CREDIT_CARD"]

    pipe = Pipeline([
        tasks.predictive.PIIMasking(
            pii_types=pii_types_list,
            model=batch_runtime.model,
            model_settings=batch_runtime.model_settings,
            batch_size=batch_runtime.batch_size,
        )
    ])
    docs = list(pipe(pii_masking_docs))

    assert len(docs) == 2
    for doc in docs:
        assert doc.text
        assert "PIIMasking" in doc.results


@pytest.mark.parametrize("batch_runtime", [ModelType.outlines], indirect=["batch_runtime"])
def test_evaluation(batch_runtime) -> None:
    """Test evaluation for PII masking without running pipeline."""
    task = tasks.predictive.PIIMasking(model=batch_runtime.model, task_id="pii")

    # 1. Full overlap
    doc_full = Doc(text="My email is test@example.com")
    res_full = Result(
        masked_text="My email is [MASKED]",
        pii_entities=[PIIEntity(entity_type="EMAIL", text="test@example.com", start=12, end=28)]
    )
    doc_full.results["pii"] = res_full
    doc_full.gold["pii"] = res_full
    report_full = task.evaluate([doc_full])
    assert report_full.metrics[task.metric] == 1.0

    # 2. Partial overlap
    doc_partial = Doc(text="My email is test@example.com and phone is 123456")
    # Pred has both
    res_pred = Result(
        masked_text="My email is [MASKED] and phone is [MASKED]",
        pii_entities=[
            PIIEntity(entity_type="EMAIL", text="test@example.com", start=12, end=28),
            PIIEntity(entity_type="PHONE", text="123456", start=42, end=48)
        ]
    )
    # Gold has only email
    res_gold = Result(
        masked_text="My email is [MASKED] and phone is 123456",
        pii_entities=[PIIEntity(entity_type="EMAIL", text="test@example.com", start=12, end=28)]
    )
    doc_partial.results["pii"] = res_pred
    doc_partial.gold["pii"] = res_gold
    report_partial = task.evaluate([doc_partial])
    # TP=1, FP=1, FN=0 -> Precision=0.5, Recall=1.0 -> F1=0.666...
    assert 0.6 < report_partial.metrics[task.metric] < 0.7

    # 3. No overlap
    doc_none = Doc(text="My email is test@example.com")
    res_none_pred = Result(
        masked_text="My email is [MASKED]",
        pii_entities=[PIIEntity(entity_type="EMAIL", text="test@example.com", start=12, end=28)]
    )
    res_none_gold = Result(
        masked_text="My name is [MASKED]",
        pii_entities=[PIIEntity(entity_type="PERSON", text="John", start=11, end=15)]
    )
    doc_none.results["pii"] = res_none_pred
    doc_none.gold["pii"] = res_none_gold
    report_none = task.evaluate([doc_none])
    assert report_none.metrics[task.metric] == 0.0



"""Base class for predictive tasks composed with model wrappers and bridges."""

from __future__ import annotations

import abc
import itertools
import json
import logging
import sys
import warnings
from collections.abc import Callable, Iterable, Sequence
from pathlib import Path
from typing import Any, Generic

import datasets
import dspy
import pydantic

from sieves.data import Doc
from sieves.model_wrappers import ModelType, ModelWrapperInferenceMode  # noqa: F401
from sieves.model_wrappers.types import ModelSettings, TokenUsage
from sieves.model_wrappers.utils import init_model_wrapper
from sieves.serialization import Config
from sieves.tasks import optimization
from sieves.tasks.core import Task
from sieves.tasks.distillation.types import DistillationFramework
from sieves.tasks.predictive.evaluation import TaskEvaluationReport
from sieves.tasks.predictive.schemas.core import (
    EvaluationSignature,
    FewshotExample,
    TaskBridge,
    TaskPromptSignature,
    TaskResult,
)
from sieves.tasks.types import Model


class PredictiveTask(Generic[TaskPromptSignature, TaskResult, TaskBridge], Task, abc.ABC):
    """Base class for predictive tasks."""

    # Threshold used for evaluation and any other situatoin in which we need to convert scores into a binary
    # representation.
    THRESHOLD = 0.5

    def __init__(
        self,
        model: Model,
        task_id: str | None,
        include_meta: bool,
        batch_size: int,
        overwrite: bool,
        prompt_instructions: str | None,
        fewshot_examples: Sequence[FewshotExample],
        model_settings: ModelSettings,
        condition: Callable[[Doc], bool] | None = None,
    ):
        """Initialize PredictiveTask.

        :param task_id: Task ID.
        :param model: Model to use.
        :param include_meta: Whether to include meta information generated by the task.
        :param batch_size: Batch size to use for inference. Use -1 to process all documents at once.
        :param overwrite: Some tasks, e.g. anonymization or translation, output a modified version of the input text.
            If True, these tasks overwrite the original document text. If False, the result will just be stored in the
            documents' `.results` field.
        :param prompt_instructions: Custom prompt instructions. If None, default instructions are used.
        :param fewshot_examples: Few-shot examples.
        :param model_settings: Settings for structured generation. Use the `inference_mode` field to specify the
            inference mode for the model wrapper. If not provided, the model wrapper will use its default mode.
        :param condition: Optional callable that determines whether to process each document.
        """
        super().__init__(task_id=task_id, include_meta=include_meta, batch_size=batch_size, condition=condition)

        self._model_wrapper = init_model_wrapper(model, model_settings)
        self._model_settings = model_settings
        self._overwrite = overwrite
        self._custom_prompt_instructions = prompt_instructions
        self._fewshot_examples = fewshot_examples
        self._bridge = self._init_bridge(ModelType.get_model_type(self._model_wrapper))

        self._validate_fewshot_examples()

    def _validate_fewshot_examples(self) -> None:
        """Validate few‑shot examples.

        :raises ValueError: if fewshot examples don't pass validation.
        """
        pass

    @property
    @abc.abstractmethod
    def fewshot_example_type(self) -> type[FewshotExample]:
        """Return few-shot example type.

        :return: Few-shot example type.
        """

    @abc.abstractmethod
    def _init_bridge(self, model_type: ModelType) -> TaskBridge:
        """Initialize bridge.

        :param model_type: Type of model to initialize bridge for.
        :return _TaskBridge: ModelWrapper task bridge.
        """

    @staticmethod
    @abc.abstractmethod
    def supports() -> set[ModelType]:
        """Return supported model types.

        :return set[ModelType]: Supported model types.
        """

    @property
    def prompt_template(self) -> str:
        """Return prompt template.

        :return str | None: Prompt template.
        """
        prompt_template = self._bridge.prompt_template
        assert isinstance(prompt_template, str)
        return prompt_template

    @property
    def prompt_signature_description(self) -> str | None:
        """Return prompt signature description.

        :return str | None: Prompt signature description.
        """
        assert hasattr(self._bridge, "prompt_signature_description")
        sig_desc = self._bridge.prompt_signature_description
        assert sig_desc is None or isinstance(sig_desc, str)
        return sig_desc

    def _call(self, docs: Iterable[Doc]) -> Iterable[Doc]:
        """Execute the task on a set of documents.

        :param docs: Documents to process.
        :return: Processed documents as an iterable.
        """
        # 1. Compile expected prompt signatures.
        signature = self._bridge.prompt_signature

        # 2. Build executable.
        executable = self._model_wrapper.build_executable(
            inference_mode=self._bridge.inference_mode,
            prompt_template=self.prompt_template,
            prompt_signature=signature,
            fewshot_examples=self._fewshot_examples,
        )

        # Compute batch-wise results.
        docs = iter(docs)
        batch_size = self._batch_size if self._batch_size > 0 else sys.maxsize
        while docs_batch := [doc for doc in itertools.islice(docs, batch_size)]:
            if len(docs_batch) == 0:
                break

            # 3. Extract values from docs to inject/render those into prompt templates.
            docs_values = self._bridge.extract(docs_batch)
            assert len(docs_values) == len(docs_batch)

            # 4. Map extracted docs values onto chunks.
            docs_chunks_offsets: list[tuple[int, int]] = []
            docs_chunks: list[dict[str, Any]] = []
            for doc, doc_values in zip(docs_batch, docs_values):
                assert doc.text
                doc_chunks_values = [doc_values | {"text": chunk} for chunk in (doc.chunks or [doc.text])]
                docs_chunks_offsets.append((len(docs_chunks), len(docs_chunks) + len(doc_chunks_values)))
                docs_chunks.extend(doc_chunks_values)

            # 5. Execute prompts per chunk.
            results_raw = executable(docs_chunks)
            assert len(results_raw) == len(docs_chunks)

            # Extract metadata.
            results, raw_outputs, usages = (list(t) for t in zip(*results_raw))

            # 6. Consolidate chunk results.
            results = self._bridge.consolidate(results, docs_chunks_offsets)
            assert len(results) == len(docs_batch)

            # 7. Integrate results into docs.
            docs_batch = self._bridge.integrate(results, docs_batch)

            # 8. Integrate raw outputs and token usage into docs.
            self._integrate_usage(docs_batch, usages, docs_chunks_offsets, raw_outputs)

            yield from docs_batch

    def _integrate_usage(
        self,
        docs_batch: list[Doc],
        usages: Sequence[TokenUsage],
        docs_chunks_offsets: list[tuple[int, int]],
        raw_outputs: Sequence[Any],
    ) -> None:
        """Integrate raw outputs and token usage into document instances.

        :param docs_batch: Current batch of documents.
        :param usages: Token usage per chunk.
        :param docs_chunks_offsets: Mapping of chunks to documents.
        :param raw_outputs: Raw model outputs per chunk.
        """
        for i, (start, end) in enumerate(docs_chunks_offsets):
            doc = docs_batch[i]
            doc_usages = usages[start:end]

            # Aggregate usage for this task.
            input_tokens_list = [u.input_tokens for u in doc_usages if u.input_tokens is not None]
            output_tokens_list = [u.output_tokens for u in doc_usages if u.output_tokens is not None]

            task_input_tokens = sum(input_tokens_list) if input_tokens_list else None
            task_output_tokens = sum(output_tokens_list) if output_tokens_list else None

            if self._include_meta:
                doc.meta[self._task_id] = {
                    "raw": raw_outputs[start:end],
                    "usage": {
                        "input_tokens": task_input_tokens,
                        "output_tokens": task_output_tokens,
                        # Include per-chunk usage in raw meta.
                        "chunks": [u.model_dump() for u in doc_usages],
                    },
                }

            # Update top-level usage.
            usage = doc.meta.get("usage", {"input_tokens": None, "output_tokens": None})
            if task_input_tokens is not None:
                current_input = usage.get("input_tokens")
                usage["input_tokens"] = (current_input or 0) + task_input_tokens
            if task_output_tokens is not None:
                current_output = usage.get("output_tokens")
                usage["output_tokens"] = (current_output or 0) + task_output_tokens

            doc.meta["usage"] = usage

    @property
    def fewshot_examples(self) -> Sequence[FewshotExample]:
        """Return few-shot examples.

        :return: Few-shot examples.
        """
        return self._fewshot_examples

    @property
    def _state(self) -> dict[str, Any]:
        return {
            **super()._state,
            "model": self._model_wrapper.model,
            "model_settings": self._model_wrapper.model_settings.model_dump(),
            "prompt_instructions": self._custom_prompt_instructions,
            "fewshot_examples": self._fewshot_examples,
        }

    @classmethod
    def deserialize(
        cls, config: Config, **kwargs: dict[str, Any]
    ) -> PredictiveTask[TaskPromptSignature, TaskResult, TaskBridge]:
        """Generate PredictiveTask instance from config.

        :param config: Config to generate instance from.
        :param kwargs: Values to inject into loaded config.
        :return PredictiveTask[TaskPromptSignature, TaskResult, _TaskBridge]: Deserialized PredictiveTask instance.
        """
        init_dict = config.to_init_dict(cls, **kwargs)
        init_dict["model_settings"] = ModelSettings.model_validate(init_dict["model_settings"])

        return cls(**init_dict)

    @property
    @abc.abstractmethod
    def prompt_signature(self) -> type[pydantic.BaseModel]:
        """Return the Pydantic prompt signature for this task.

        :return: Pydantic model class.
        """

    @property
    def metric(self) -> str:
        """Return metric name.

        :return str: Metric name.
        """
        return "Similarity (LLM-judged)"

    def evaluate(
        self, docs: Iterable[Doc], judge: dspy.LM | None = None, failure_threshold: float = 0.5
    ) -> TaskEvaluationReport:
        """Evaluate task performance using DSPy-based evaluation.

        :param docs: Documents to evaluate.
        :param judge: Optional DSPy LM instance to use as judge for generative tasks.
        :param failure_threshold: Decision threshold for whether to mark predicitions as failures.
        :return: Evaluation report.
        """
        truths: list[Any] = []
        preds: list[Any] = []
        failures: list[Doc] = []

        # Evaluate each doc individually to identify failed predictions.
        for doc in docs:
            if self.id not in doc.results:
                continue

            pred = doc.results[self.id]
            gold = doc.gold.get(self.id, None)

            # Accumulate for corpus-level metrics.
            truths.append(gold)
            preds.append(pred)

            # If gold or prediction is None: we cannot do proper evalution, so we just check whether they're both None
            # to compute score for failure analysis.
            if gold is None or pred is None:
                if gold is not None or pred is not None:
                    failures.append(doc)
            else:
                # Convert result and gold to DSPy representation.
                truth = dspy.Example(**self._task_result_to_dspy_dict(gold))
                pred_dspy = dspy.Prediction(**self._task_result_to_dspy_dict(pred))

                # Call internal evaluation logic for per-doc failure analysis.
                score = self._evaluate_dspy_example(truth, pred_dspy, trace=None, model=judge)

                if score < failure_threshold:
                    failures.append(doc)

        # Evaluate on corpus level to obtain representative metrics.
        metrics = self._compute_metrics(truths, preds, judge=judge)

        return TaskEvaluationReport(
            metrics=metrics,
            task_id=self.id,
            failures=failures,
        )

    def _compute_metrics(self, truths: list[Any], preds: list[Any], judge: dspy.LM | None = None) -> dict[str, float]:
        """Compute corpus-level metrics.

        By default, this computes the average of `_evaluate_dspy_example` over all documents.

        :param truths: List of ground truths.
        :param preds: List of predictions.
        :param judge: Optional DSPy LM instance to use as judge for generative tasks.
        :return: Dictionary of metrics.
        """
        scores: list[float] = []
        for gold, pred in zip(truths, preds):
            if gold is None or pred is None:
                scores.append(1.0 if gold is None and pred is None else 0.0)
            else:
                truth = dspy.Example(**self._task_result_to_dspy_dict(gold))
                pred_dspy = dspy.Prediction(**self._task_result_to_dspy_dict(pred))
                scores.append(self._evaluate_dspy_example(truth, pred_dspy, trace=None, model=judge))

        avg_score = sum(scores) / len(scores) if scores else 0.0
        return {self.metric: avg_score}

    def _task_result_to_dspy_dict(self, result: Any) -> dict[str, Any]:
        """Convert a task result or gold data to a dictionary suitable for DSPy evaluation.

        :param result: Task result or gold data.
        :return: Dictionary representation.
        """
        if hasattr(result, "model_dump"):
            return result.model_dump(serialize_as_any=True)
        if isinstance(result, dict):
            return result
        return {"output": result}

    @abc.abstractmethod
    def to_hf_dataset(self, docs: Iterable[Doc], threshold: float | None = None) -> datasets.Dataset:
        """Create Hugging Face datasets.Dataset from docs.

        :param docs: Docs to convert.
        :param threshold: Threshold to apply when converting logits/confidence scores into labels or other structured
            predictions. Defaults to `THRESHOLD`.
        :return datasets.Dataset: Hugging Face dataset.
        """

    @abc.abstractmethod
    def distill(
        self,
        base_model_id: str,
        framework: DistillationFramework,
        data: datasets.Dataset | Sequence[Doc],
        output_path: Path | str,
        val_frac: float,
        init_kwargs: dict[str, Any] | None = None,
        train_kwargs: dict[str, Any] | None = None,
        seed: int | None = None,
    ) -> None:
        """Distill a model for this task.

        Doc instances must have `.results[task_id]` - otherwise this terminates with an error.

        This method fine-tunes a base model using distillation techniques based on the provided framework. It splits
        the input dataset, trains the model, and saves the resulting model and metadata to the specified output path.

        :param base_model_id: ID of Hugging Face model to use as base for distillation. The chosen model will be
            fine-tuned on the target task's results.
        :param framework: Which distillation framework to use.
        :param data: Docs to extract results from.
        :param output_path: Path to store distilled model and training metadata at.
        :param val_frac: Fraction of data to use for validation set.
        :param seed: RNG seed.
        :param init_kwargs: Kwargs passed on to model/trainer initialization.
        :param train_kwargs: Kwargs passed on to training _call.
        :raises KeyError: If expected columns don't exist in `hf_dataset`.
        """

    @staticmethod
    def _split_dataset(
        hf_dataset: datasets.Dataset, train_frac: float, val_frac: float, seed: int | None
    ) -> datasets.DatasetDict:
        """Split dataset.

        :param hf_dataset: Dataset to split.
        :param train_frac: Fractions for training set. `train_frac` + `val_frac` must sum up to 1.
        :param val_frac: Fractions for validation set. `train_frac` + `val_frac` must sum up to 1.
        :param seed: RNG seed.
        :return: Train, val sets; mapping of rows to sets.
        :raises ValueError: If fractions don't sum up to 1.
        """
        if not 0 < val_frac < 1:
            raise ValueError(f"`val_frac` must be greater than 0 and less than 1, but got {val_frac}.")
        if not abs(train_frac + val_frac - 1.0) < 1e-9:
            raise ValueError(f"Split fractions must sum to 1.0, but got {train_frac}, {val_frac}.")

        if val_frac > 0:
            train_val_dataset = hf_dataset.train_test_split(test_size=val_frac, shuffle=True, seed=seed)
            return datasets.DatasetDict({"train": train_val_dataset["train"], "val": train_val_dataset["test"]})

        return datasets.DatasetDict({"train": hf_dataset})

    def _get_task_signature(self) -> type[dspy.Signature] | type[dspy.Module]:
        """Get DSPy signature for this task.

        By default this uses the task signature of the DSPy bridge for this task. If none is found, this fails. Can be
        overwritten with a custom task signature, if no DSPy bridge will be implemented for this task.

        :return: DSPy signature for this task.
        :raises KeyError: If no DSPy bridge defined for this task.
        """
        try:
            dspy_bridge = self._bridge if self._model_wrapper == ModelType.dspy else self._init_bridge(ModelType.dspy)
            assert hasattr(dspy_bridge, "prompt_signature")
            assert issubclass(dspy_bridge.prompt_signature, dspy.Signature | dspy.Module)
            return dspy_bridge.prompt_signature

        except KeyError as err:
            raise KeyError(f"DSPy bridge not available for task {self.__class__.__name__}.") from err

    def _evaluate_dspy_example(self, truth: dspy.Example, pred: dspy.Prediction, trace: Any, model: dspy.LM) -> float:
        """Evaluate DSPy example for optimization or evaluation.

        By default this implements an LLM-based evaluation that uses the model that drives optimization to compare
        results. Where possible this should be replaced by a targeted, deterministic evaluation.

        :param truth: Ground truth.
        :param pred: Predicted value.
        :param model: Model used by optimizer.
        :param trace: Optional trace information.
        :return: Metric value between 0.0 and 1.0.
        :raises KeyError: If target fields are missing from truth or prediction.
        :raises ValueError: If similarity score cannot be parsed from LLM response or no judge model provided.
        """
        if model is None:
            raise ValueError(
                f"No judge model provided for task {self.__class__.__name__}. "
                "Generative tasks require a DSPy LM judge for evaluation."
            )

        if self._fewshot_examples:
            target_fields = list(self._fewshot_examples[0].target_fields)
        else:
            # Fallback to all fields in truth that are not input fields.
            target_fields = [k for k in truth.keys() if k not in ("text",)]

        # Filter truth and pred to only include target fields.
        truth_filtered: dict[str, Any] = {}
        for field in target_fields:
            if field not in truth:
                raise KeyError(f"Target field '{field}' missing from ground truth example: {dict(truth)}.")
            truth_filtered[field] = truth[field]

        pred_filtered: dict[str, Any] = {}
        for field in target_fields:
            if field not in pred:
                raise KeyError(f"Target field '{field}' missing from prediction: {dict(pred)}.")
            pred_filtered[field] = pred[field]

        # Serialize to readable format
        target_fields_str = ", ".join(target_fields)
        ground_truth_str = json.dumps(truth_filtered, indent=2, default=str)
        prediction_str = json.dumps(pred_filtered, indent=2, default=str)

        # Create DSPy evaluator module
        evaluator = dspy.Predict(EvaluationSignature)

        # Call evaluator with the model
        with dspy.context(lm=model):
            result = evaluator(
                target_fields=target_fields_str,
                ground_truth=ground_truth_str,
                prediction=prediction_str,
            )

        # Parse and validate score
        try:
            score = float(result.similarity_score)
        except (ValueError, AttributeError, TypeError) as err:
            raise ValueError(
                f"Failed to parse similarity score from LLM response. Got: {getattr(result, 'similarity_score', None)}"
            ) from err

        # Clamp to [0, 1].
        return max(0.0, min(1.0, score))

    def optimize(self, optimizer: optimization.Optimizer, verbose: bool = True) -> tuple[str, Sequence[FewshotExample]]:
        """Optimize task prompt and few-shot examples with the available optimization config.

        Updates task to use best prompt and few-shot examples found by the optimizer.

        :param optimizer: Optimizer to run.
        :param verbose: Whether to suppress output. DSPy produces a good amount of logs, so this can be useful to
            not pollute your terminal. Only warnings and errors will be printed.

        :return tuple[str, Sequence[FewshotExample]]: Best found prompt and few-shot examples.
        """
        assert len(self._fewshot_examples) > 1, "At least two few-shot examples need to be provided to optimize."

        # Run optimizer to get best prompt and few-shot examples.
        signature = self._get_task_signature()
        dspy_examples = [ex.to_dspy() for ex in self._fewshot_examples]

        def _pred_eval(truth: dspy.Example, pred: dspy.Prediction, trace: Any | None = None) -> float:
            """Wrap optimization evaluation, inject model.

            :param truth: Ground truth.
            :param pred: Predicted value.
            :param trace: Optional trace information.
            :return: Metric value between 0.0 and 1.0.
            :raises KeyError: If target fields are missing from truth or prediction.
            :raises ValueError: If similarity score cannot be parsed from LLM response.
            """
            return self._evaluate_dspy_example(truth, pred, trace, model=optimizer.model)

        if verbose:
            best_prompt, best_examples = optimizer(signature, dspy_examples, _pred_eval, verbose=verbose)
        else:
            # Temporarily suppress DSPy logs.
            dspy_logger = logging.getLogger("dspy")
            optuna_logger = logging.getLogger("optuna")
            original_dspy_level = dspy_logger.level
            original_optuna_level = optuna_logger.level

            try:
                dspy_logger.setLevel(logging.ERROR)
                optuna_logger.setLevel(logging.ERROR)
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    best_prompt, best_examples = optimizer(signature, dspy_examples, _pred_eval, verbose=verbose)
            finally:
                dspy_logger.setLevel(original_dspy_level)
                optuna_logger.setLevel(original_optuna_level)

        # Update few-shot examples and prompt instructions.
        fewshot_example_cls = self._fewshot_examples[0].__class__
        self._fewshot_examples = [fewshot_example_cls.from_dspy(ex) for ex in best_examples]
        self._validate_fewshot_examples()
        self._custom_prompt_instructions = best_prompt

        # Reinitialize bridge to use new prompt and few-shot examples.
        self._bridge = self._init_bridge(ModelType.get_model_type(self._model_wrapper))

        return best_prompt, self._fewshot_examples



"""Summarization task."""

from sieves.tasks.predictive.schemas.summarization import FewshotExample, Result
from sieves.tasks.predictive.summarization.core import Summarization

__all__ = ["Summarization", "FewshotExample", "Result"]



"""LangChain model wrapper for structured outputs using Pydantic."""

import asyncio
import enum
from collections.abc import Iterable, Sequence
from typing import Any, override

import langchain_core.language_models
import nest_asyncio
import pydantic

from sieves.model_wrappers.core import Executable, PydanticModelWrapper
from sieves.model_wrappers.types import TokenUsage

nest_asyncio.apply()

Model = langchain_core.language_models.BaseChatModel
PromptSignature = pydantic.BaseModel
Result = pydantic.BaseModel


class InferenceMode(enum.Enum):
    """Available inference modes."""

    structured = "structured"


class LangChain(PydanticModelWrapper[PromptSignature, Result, Model, InferenceMode]):
    """ModelWrapper for LangChain."""

    @override
    @property
    def inference_modes(self) -> type[InferenceMode]:
        return InferenceMode

    @override
    def build_executable(
        self,
        inference_mode: InferenceMode,
        prompt_template: str | None,  # noqa: UP007
        prompt_signature: type[PromptSignature] | PromptSignature,
        fewshot_examples: Sequence[pydantic.BaseModel] = tuple(),
    ) -> Executable[Result | None]:
        assert isinstance(prompt_signature, type)
        cls_name = self.__class__.__name__
        template = self._create_template(prompt_template)
        model = self._model.with_structured_output(prompt_signature, include_raw=True)

        def execute(values: Sequence[dict[str, Any]]) -> Sequence[tuple[Result | None, Any, TokenUsage]]:
            """Execute prompts with model wrapper for given values.

            :param values: Values to inject into prompts.
            :return: Sequence of tuples containing results, raw outputs, and token usage. Results are None if
                corresponding prompt failed.
            """
            match inference_mode:
                case InferenceMode.structured:

                    def generate(prompts: list[str]) -> Iterable[tuple[Result, Any, TokenUsage]]:
                        try:
                            results = asyncio.run(model.abatch(prompts, **self._inference_kwargs))
                            for res in results:
                                usage = TokenUsage()
                                raw = res["raw"]
                                if hasattr(raw, "usage_metadata") and raw.usage_metadata:
                                    usage.input_tokens = raw.usage_metadata.get("input_tokens")
                                    usage.output_tokens = raw.usage_metadata.get("output_tokens")

                                yield res["parsed"], raw, usage

                        except Exception as err:
                            raise RuntimeError(
                                f"Encountered problem in parsing {cls_name} output. Double-check your prompts and "
                                f"examples."
                            ) from err

                    generator = generate
                case _:
                    raise ValueError(f"Inference mode {inference_mode} not supported by {cls_name} model wrapper.")

            return self._infer(generator, template, values)

        return execute



"""
Test file containing examples for the Preprocessing guide.

These code blocks are referenced in docs/guides/preprocessing.md using snippet injection.

Usage in markdown:
    ```python
    --8<-- "sieves/tests/docs/test_preprocessing.py:ingestion-basic"
    ```
"""

import pytest


def test_basic_ingestion_example():
    """Test basic document ingestion/parsing example."""
    # --8<-- [start:ingestion-basic]
    from sieves import Pipeline, tasks, Doc

    # Create a document parser
    parser = tasks.preprocessing.Ingestion()

    # Create a pipeline with the parser
    pipeline = Pipeline([parser])

    # Process documents (requires actual PDF/DOCX files)
    docs = [
        Doc(uri="path/to/document.pdf"),
        Doc(uri="path/to/another.docx")
    ]
    # Note: Ingestion requires actual files and optional dependencies
    # Install with: pip install "sieves[ingestion]"
    # --8<-- [end:ingestion-basic]

    # Verify setup is correct
    assert parser is not None
    assert pipeline is not None
    assert len(docs) == 2


def test_custom_converter_example():
    """Test ingestion with custom converter and export format."""
    # --8<-- [start:ingestion-custom-converter]
    from sieves import Pipeline, tasks, Doc

    # Create a document parser with custom export format
    parser = tasks.preprocessing.Ingestion(export_format="html")

    # Create a pipeline with the parser
    pipeline = Pipeline([parser])

    # Process documents (requires actual PDF/DOCX files)
    docs = [
        Doc(uri="path/to/document.pdf"),
        Doc(uri="path/to/another.docx")
    ]
    # Note: Ingestion requires actual files and optional dependencies
    # Install with: pip install "sieves[ingestion]"
    # --8<-- [end:ingestion-custom-converter]

    assert parser is not None
    assert pipeline is not None
    assert len(docs) == 2


def test_chunking_example(example_tokenizer):
    """Test Chunking task with chonkie library."""
    tokenizer = example_tokenizer  # For testing, use fixture

    # --8<-- [start:chunking-chonkie-basic]
    import chonkie
    import tokenizers
    from sieves import Pipeline, tasks, Doc

    # Create a tokenizer for chunking
    tokenizer = tokenizers.Tokenizer.from_pretrained("bert-base-uncased")

    # Create a token-based chunker
    chunker = tasks.Chunking(
        chunker=chonkie.TokenChunker(tokenizer, chunk_size=512, chunk_overlap=50)
    )

    # Create and run the pipeline
    pipeline = Pipeline(chunker)
    doc = Doc(text="Your long document text here...")
    chunked_docs = list(pipeline([doc]))

    # Access the chunks
    for chunk in chunked_docs[0].chunks:
        print(f"Chunk: {chunk}")
    # --8<-- [end:chunking-chonkie-basic]

    # Assertions for testing
    assert len(chunked_docs) > 0
    assert chunked_docs[0].chunks is not None


def test_combined_preprocessing_pipeline(example_tokenizer):
    """Test combined preprocessing pipeline with chunking."""
    tokenizer = example_tokenizer  # For testing, use fixture

    # --8<-- [start:preprocessing-combined-pipeline]
    from sieves import tasks, Doc, Pipeline
    import chonkie
    import tokenizers

    # Create a tokenizer and chunker
    tokenizer = tokenizers.Tokenizer.from_pretrained("bert-base-uncased")
    chunker = tasks.Chunking(
        chunker=chonkie.TokenChunker(tokenizer, chunk_size=512, chunk_overlap=50)
    )

    # Create a pipeline
    pipeline = Pipeline(chunker)

    # Process a document with text
    doc = Doc(text="This is a long document that will be split into chunks. " * 100)
    processed_doc = list(pipeline([doc]))[0]

    # Access the chunks
    print(f"Number of chunks: {len(processed_doc.chunks)}")
    for i, chunk in enumerate(processed_doc.chunks):
        print(f"Chunk {i}: {chunk[:100]}...")  # Print first 100 chars of each chunk
    # --8<-- [end:preprocessing-combined-pipeline]

    assert processed_doc.chunks is not None
    assert len(processed_doc.chunks) > 0


def test_metadata_inclusion_example():
    """Test metadata inclusion in preprocessing tasks."""
    # --8<-- [start:metadata-inclusion]
    from sieves import tasks

    parser = tasks.preprocessing.Ingestion(include_meta=True)
    # --8<-- [end:metadata-inclusion]

    # Assertion for testing
    assert parser is not None


def test_metadata_access_example():
    """Demonstrate how to access metadata from processed documents."""
    # Create dummy data to demonstrate the pattern
    from sieves import Doc

    # Simulate processed docs with metadata
    processed_docs = [
        Doc(text="example", meta={"Ingestion": {"status": "ok"}, "Chunker": {"num_chunks": 3}})
    ]

    # --8<-- [start:metadata-access]
    doc = processed_docs[0]
    print(doc.meta["Ingestion"])  # Access parser metadata
    print(doc.meta["Chunker"])  # Access chunker metadata
    # --8<-- [end:metadata-access]

    assert doc.meta is not None



"""Schemas for relation extraction task."""

from __future__ import annotations

import dspy
import gliner2.inference.engine
import pydantic

from sieves.model_wrappers import dspy_, gliner_, langchain_, outlines_
from sieves.tasks.predictive.schemas.core import FewshotExample as BaseFewshotExample


class RelationEntity(pydantic.BaseModel, frozen=True):
    """Entity involved in a relation.

    Attributes:
        text: Surface text of the entity.
        entity_type: Type of the entity.
    """

    text: str
    entity_type: str


class RelationTriplet(pydantic.BaseModel, frozen=True):
    """Triplet representing a relation between two entities.

    Attributes:
        head: The subject entity.
        relation: The type of relation.
        tail: The object entity.
        score: Confidence score.
    """

    head: RelationEntity
    relation: str
    tail: RelationEntity
    score: float | None = None


# --8<-- [start:Result]
class Result(pydantic.BaseModel):
    """Result of a relation extraction task.

    Attributes:
        triplets: List of extracted relation triplets.
    """

    triplets: list[RelationTriplet]


# --8<-- [end:Result]


class RelationEntityWithContext(pydantic.BaseModel):
    """Entity mention with text span, type, and context for span discovery.

    Attributes:
        text: Surface text of the entity.
        context: Short context around the entity.
        entity_type: Type of the entity.
    """

    text: str
    context: str
    entity_type: str


class RelationTripletWithContext(pydantic.BaseModel):
    """Triplet with context for span discovery.

    Attributes:
        head: The head entity with context.
        relation: The relation type.
        tail: The tail entity with context.
        score: Confidence score.
    """

    head: RelationEntityWithContext
    relation: str
    tail: RelationEntityWithContext
    score: float | None = None


class FewshotExample(BaseFewshotExample):
    """Few-shot example for relation extraction.

    Attributes:
        text: Input text.
        triplets: Expected relation triplets.
    """

    triplets: list[RelationTriplet]

    @property
    def target_fields(self) -> tuple[str, ...]:
        """Return target fields.

        :return: Target fields.
        """
        return ("triplets",)


TaskModel = dspy_.Model | gliner_.Model | langchain_.Model | outlines_.Model
TaskPromptSignature = (
    type[dspy.Signature]
    | type[pydantic.BaseModel]
    | gliner2.inference.engine.Schema
    | gliner2.inference.engine.StructureBuilder
)
TaskResult = Result



"""
Test file containing examples for the Serialization guide.

These code blocks are referenced in docs/guides/serialization.md using snippet injection.
"""

import pytest
from pathlib import Path


def test_basic_serialization(small_outlines_model, tmp_path):
    """Test basic pipeline serialization and loading."""
    model = small_outlines_model

    # --8<-- [start:serialization-basic-pipeline]
    import outlines
    from transformers import AutoModelForCausalLM, AutoTokenizer
    from sieves import Pipeline, tasks, Doc
    from pathlib import Path

    # Create a basic classification pipeline
    model_name = "HuggingFaceTB/SmolLM-135M-Instruct"
    model = outlines.models.from_transformers(
        AutoModelForCausalLM.from_pretrained(model_name),
        AutoTokenizer.from_pretrained(model_name)
    )
    classifier = tasks.predictive.Classification(labels=["science", "politics"], model=model)
    pipeline = Pipeline([classifier])

    # Save the pipeline configuration
    config_path = Path("classification_pipeline.yml")
    pipeline.dump(config_path)

    # Load the pipeline configuration
    loaded_pipeline = Pipeline.load(config_path, [{"model": outlines.models.from_transformers(
        AutoModelForCausalLM.from_pretrained(model_name),
        AutoTokenizer.from_pretrained(model_name)
    )}])

    # Use the loaded pipeline
    doc = Doc(text="Special relativity applies to all physical phenomena in the absence of gravity.")
    results = list(loaded_pipeline([doc]))
    print(results[0].results["Classification"])
    # --8<-- [end:serialization-basic-pipeline]

    # Assertions
    assert results[0].results["Classification"] is not None

    # Cleanup
    config_path.unlink()


def test_complex_serialization(example_tokenizer, small_outlines_model, tmp_path):
    """Test serialization with complex third-party objects."""
    tokenizer = example_tokenizer
    model = small_outlines_model

    # --8<-- [start:serialization-complex-pipeline]
    # --8<-- [start:serialization-complex-setup]
    import chonkie
    import tokenizers
    import outlines
    import pydantic
    from transformers import AutoModelForCausalLM, AutoTokenizer
    from sieves import Pipeline, tasks

    # Create a tokenizer for chunking
    tokenizer = tokenizers.Tokenizer.from_pretrained("bert-base-uncased")
    chonkie_chunker = chonkie.TokenChunker(tokenizer, chunk_size=512, chunk_overlap=50)
    chunker = tasks.Chunking(chunker=chonkie_chunker)

    model_name = "HuggingFaceTB/SmolLM-135M-Instruct"
    model = outlines.models.from_transformers(
        AutoModelForCausalLM.from_pretrained(model_name),
        AutoTokenizer.from_pretrained(model_name)
    )
    # --8<-- [end:serialization-complex-setup]


    # --8<-- [start:serialization-complex-entity-task]
    class PersonInfo(pydantic.BaseModel, frozen=True):
        name: str
        age: int | None = None
        occupation: str | None = None


    extractor = tasks.predictive.InformationExtraction(entity_type=PersonInfo, model=model)
    # --8<-- [end:serialization-complex-entity-task]

    # --8<-- [start:serialization-complex-save]
    # Create and save the pipeline
    pipeline = chunker + extractor
    pipeline.dump("extraction_pipeline.yml")
    # --8<-- [end:serialization-complex-save]

    # --8<-- [start:serialization-complex-load]
    # Load the pipeline with initialization parameters for each task
    loaded_pipeline = Pipeline.load(
        "extraction_pipeline.yml",
        [
            {"chunker": chonkie_chunker},
            {
                "entity_type": PersonInfo,
                "model": model
            },
        ]
    )
    # --8<-- [end:serialization-complex-load]
    # --8<-- [end:serialization-complex-pipeline]

    # Assertions
    assert loaded_pipeline is not None
    assert len(loaded_pipeline.tasks) == 2

    # Cleanup
    Path("extraction_pipeline.yml").unlink()



"""Ingestion task implementation."""

from __future__ import annotations

from collections.abc import Callable, Iterable
from types import NoneType
from typing import Any, Literal

from sieves.data.doc import Doc
from sieves.serialization import Config
from sieves.tasks.core import Task
from sieves.tasks.preprocessing.ingestion import docling_, marker_

Converter = docling_.Converter | marker_.Converter


class Ingestion(Task):
    """Base class for Ingestion tasks that extract text from documents.

    This unified interface allows different Ingestion converters to be used interchangeably.
    """

    def __init__(
        self,
        converter: Converter | None = None,
        export_format: Literal["markdown", "html"] = "markdown",
        task_id: str | None = None,
        include_meta: bool = False,
        batch_size: int = -1,
        condition: Callable[[Doc], bool] | None = None,
        **kwargs: Any,
    ):
        """Initialize the Ingestion task.

        :param converter: The Ingestion converter to use. If None, tries to initialize in this order:
            - A `docling` converter with default values
            - A `marker` converter with default values
            If neither is possible due to a lack of installed requirements, raises a `ModuleNotFoundError`.
        :param task_id: Task ID.
        :param include_meta: Whether to include meta information generated by the task.
        :param batch_size: Batch size to use for processing. Use -1 to process all documents at once.
        :param condition: Optional callable that determines whether to process each document.
        :param kwargs: Additional arguments for specific `Ingestion` implementations.
        """
        super().__init__(task_id=task_id, include_meta=include_meta, batch_size=batch_size, condition=condition)
        self._export_format = export_format
        self._converter = converter
        self._kwargs = kwargs
        self._task = self._init_ingestion_task()

    def _init_ingestion_task(self) -> Task:
        """Initialize the bridge for the specific Ingestion implementation.

        :raises ValueError: On unsupported converter type.
        :return: Ingestion bridge implementation.
        """
        converter_type = type(self._converter)

        # Set converter type manually in predefined order, depending on which dependencies are installed.
        if self._converter is None:
            if docling_ is not None:
                converter_type = docling_.Converter
            elif marker_ is not None:
                converter_type = marker_.Converter.__args__[0]
            else:
                raise ModuleNotFoundError(
                    "None of the supported ingestion libraries (`docling`, `marker`) are installed. Install at least "
                    "one of them to run ingestion tasks."
                )
        assert converter_type is not NoneType

        # Identify the ingestion task that maps to the specified converter's type.
        converter_module_map = {
            docling_: getattr(docling_, "Docling", None),
            marker_: getattr(marker_, "Marker", None),
        }

        for module, ingestion_task_type in converter_module_map.items():
            if ingestion_task_type is None:
                continue

            assert hasattr(module, "Converter")
            try:
                module_converter_types = module.Converter.__args__
            except AttributeError:
                module_converter_types = (module.Converter,)

            if any(issubclass(converter_type, module_model_type) for module_model_type in module_converter_types):
                ingestion_task = ingestion_task_type(
                    converter=self._converter,
                    export_format=self._export_format,
                    task_id=self.id,
                    include_meta=self._include_meta,
                    batch_size=self._batch_size,
                    **self._kwargs,
                )
                assert isinstance(ingestion_task, Task)

                return ingestion_task

        raise ValueError(
            f"converter type {self._converter} is not supported. Please check the documentation and ensure that (1) "
            f"you're providing a supported converter type and that (2) the corresponding library is installed in your "
            f"environment."
        )

    def _call(self, docs: Iterable[Doc]) -> Iterable[Doc]:
        """Process documents with Ingestion to extract text.

        :param docs: Documents to process.
        :return: Processed documents with extracted text.
        """
        docs = list(docs)
        assert all(doc.uri for doc in docs), ValueError("Documents have to have a value for .uri.")
        result = self._task(docs)

        yield from result

    @property
    def _state(self) -> dict[str, Any]:
        """Returns attributes to serialize.

        :return: Dict of attributes to serialize.
        """
        return {
            **super()._state,
            "converter": self._converter,
            "export_format": self._export_format,
            **self._kwargs,
        }

    @classmethod
    def deserialize(cls, config: Config, **kwargs: dict[str, Any]) -> Ingestion:
        """
        Generate Ingestion instance from config.

        :param config: Config to generate instance from.
        :param kwargs: Values to inject into loaded config.
        :return: Deserialized Ingestion instance.
        """
        return cls(**config.to_init_dict(cls, **kwargs))
